# --------------------------------------------------------------------------------------
# vae.py
#
# from the mlcolvar repository
#     https://github.com/mlcolvar/mlcolvar
#     Copyright (c) 2023 Luigi Bonati, Enrico Trizio, Andrea Rizzi & Michele Parrinello
#     Licensed under the MIT License (see project LICENSE file for full text)
# --------------------------------------------------------------------------------------

from typing import Optional, Tuple
import torch
import lightning.pytorch as pl
from mlcolvar.cvs import BaseCV
from biobb_pytorch.mdae.models.nn.feedforward import FeedForward
from biobb_pytorch.mdae.featurization.normalization import Normalization
from mlcolvar.core.transform.utils import Inverse
from biobb_pytorch.mdae.loss import ELBOGaussiansLoss

__all__ = ["VariationalAutoEncoder"]


class VariationalAutoEncoder(BaseCV, pl.LightningModule):

    """Variational AutoEncoder Collective Variable.

    At training time, the encoder outputs a mean and a variance for each CV
    defining a Gaussian distribution associated to the input. One sample is
    drawn from this Gaussian, and it goes through the decoder. Then the ELBO
    loss is minimized. The ELBO sums the MSE of the reconstruction and the KL
    divergence between the generated Gaussian and a N(0, 1) Gaussian.

    At evaluation time, the encoder's output mean is used as the CV, while the
    variance output and the decoder are ignored.

    **Data**: for training, it requires a DictDataset with the key ``'data'`` and
    optionally ``'weights'``. If a 'target' key is present this will be used as reference
    for the output of the decoder, otherway this will be compared with the input 'data'.
    This feature can be used to train (variational) time-lagged autoencoders like in [1]_.

    **Loss**: Evidence Lower BOund (ELBO)

    References
    ----------
    .. [1] C. X. Hernández, H. K. Wayment-Steele, M. M. Sultan, B. E. Husic, and V. S. Pande,
        “Variational encoding of complex dynamics,” Physical Review E 97, 062412 (2018).

    See also
    --------
    mlcolvar.core.loss.ELBOLoss
        Evidence Lower BOund loss function
    """

    BLOCKS = ["norm_in", "encoder", "decoder"]

    def __init__(
        self,
        n_features: int,
        n_cvs: int,
        encoder_layers: list,
        decoder_layers: Optional[list] = None,
        options: Optional[dict] = None,
        **kwargs,
    ):
        """
        Variational autoencoder constructor. Initializes two neural network modules
        (encoder and decoder). By default a module standardizing the inputs is also used.

        Parameters
        ----------
        n_cvs : int
            The dimension of the CV or, equivalently, the dimension of the latent
            space of the autoencoder.
        encoder_layers : list
            Number of neurons per layer of the encoder up to the last hidden layer.
            The size of the output layer is instead specified with ``n_cvs``
        decoder_layers : list, optional
            Number of neurons per layer of the decoder, except for the input layer
            which is specified by ``n_cvs``. If ``None`` (default), it takes automatically
            the reversed architecture of the encoder.
        options : dict[str, Any], optional
            Options for the building blocks of the model, by default ``None``.
            Available blocks are: ``'norm_in'``, ``'encoder'``, and ``'decoder'``.
            Set ``'block_name' = None`` or ``False`` to turn off a block. Encoder
            and decoder cannot be turned off.
        """
        super().__init__(in_features=n_features, out_features=n_cvs, **kwargs)

        # =======   LOSS  =======
        # ELBO loss function when latent space and reconstruction distributions are Gaussians.
        self.loss_fn = ELBOGaussiansLoss()

        # ======= OPTIONS =======
        # parse and sanitize
        options = self.parse_options(options)

        # if decoder is not given reverse the encoder
        if decoder_layers is None:
            decoder_layers = encoder_layers[::-1]

        # ======= BLOCKS =======

        # initialize norm_in
        o = "norm_in"
        if (options[o] is not False) and (options[o] is not None):
            self.norm_in = Normalization(self.in_features, **options[o])

        # initialize encoder
        # The encoder outputs two values for each CV representing mean and std.
        o = "encoder"

        # Note: The FeedForward implementing the encoder by default needs to have also
        # the nonlinearity (and eventually dropout/batchnorm) also for the output
        # layer since we'll have two separate linear layers for the mean and variance.
        if "last_layer_activation" not in options[o]:
            options[o]["last_layer_activation"] = True

        self.encoder = FeedForward([n_features] + encoder_layers, **options[o])
        self.mean_nn = torch.nn.Linear(
            in_features=encoder_layers[-1], out_features=n_cvs
        )
        self.log_var_nn = torch.nn.Linear(
            in_features=encoder_layers[-1], out_features=n_cvs
        )

        # initialize encoder
        o = "decoder"
        self.decoder = FeedForward([n_cvs] + decoder_layers + [n_features], **options[o])

        self.eval_variables = ["xhat", "z", "z_mean", "z_logvar"]

    def forward_cv(self, x: torch.Tensor) -> torch.Tensor:
        """Compute the value of the CV from preprocessed input.

        Return the mean output (ignoring the variance output) of the encoder
        after (optionally) applying the normalization to the input.

        Parameters
        ----------
        x : torch.Tensor
            Shape ``(n_batches, n_descriptors)`` or ``(n_descriptors,)``. The
            input descriptors of the CV after preprocessing.

        Returns
        -------
        cv : torch.Tensor
            Shape ``(n_batches, n_cvs)``. The CVs, i.e., the mean output of the
            encoder (the variance output is discarded).
        """
        if self.norm_in is not None:
            x = self.norm_in(x)
        x = self.encoder(x)

        # Take only the means and ignore the log variances.
        return self.mean_nn(x)

    def encode_decode(
        self, x: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Run a pass of encoding + decoding.

        The function applies the normalizing to the inputs and its reverse on
        the output.

        Parameters
        ----------
        x : torch.Tensor
            Shape ``(n_batches, n_descriptors)`` or ``(n_descriptors,)``. The
            input descriptors of the CV after preprocessing.

        Returns
        -------
        mean : torch.Tensor
            Shape ``(n_batches, n_cvs)`` of ``(n_cvs,)``. The mean of the
            Gaussian distribution associated to the input in latent space.
        log_variance : torch.Tensor
            Shape ``(n_batches, n_cvs)`` of ``(n_cvs,)``. The logarithm of the
            variance of the Gaussian distribution associated to the input in
            latent space.
        x_hat : torch.Tensor
            Shape ``(n_batches, n_descriptors)`` or ``(n_descriptors,)``. The
            reconstructed descriptors.
        """
        # Normalize inputs.
        if self.norm_in is not None:
            x = self.norm_in(x)

        # Encode input into a Gaussian distribution.
        x = self.encoder(x)
        mean, log_variance = self.mean_nn(x), self.log_var_nn(x)

        # Sample from the Gaussian distribution in latent space.
        std = torch.exp(log_variance / 2)
        z = torch.distributions.Normal(mean, std).rsample()

        # Decode sample.
        x_hat = self.decoder(z)

        # if self.norm_in is not None:
        #     x_hat = self.norm_in.inverse(x_hat)

        return z, mean, log_variance, x_hat

    def evaluate_model(self, batch, batch_idx):
        """Evaluate the model on the data, computing average loss."""

        x = batch['data']

        if 'target' in batch:
            x_ref = batch['target']
            if self.norm_in is not None:
                x_ref = self.norm_in(x_ref)
        else:
            x_ref = x

        z, mean, log_variance, x_hat = self.encode_decode(x)

        if self.norm_in is not None:
            x_hat = self.norm_in.inverse(x_hat)

        return x_hat, z, mean, log_variance

    def training_step(self, train_batch, batch_idx):
        """Single training step performed by the PyTorch Lightning Trainer."""
        x = train_batch["data"]
        loss_kwargs = {}
        if "weights" in train_batch:
            loss_kwargs["weights"] = train_batch["weights"]

        # Encode/decode.
        z, mean, log_variance, x_hat = self.encode_decode(x)

        # Reference output (compare with a 'target' key if any, otherwise with input 'data')
        if "target" in train_batch:
            x_ref = train_batch["target"]
        else:
            x_ref = x

        if self.norm_in is not None:
            x_ref = self.norm_in(x_ref)

        # Loss function.
        loss = self.loss_fn(x_ref, x_hat, mean, log_variance, **loss_kwargs)

        # Log.
        name = "train" if self.training else "valid"
        self.log(f"{name}_loss", loss, on_epoch=True, on_step=True, prog_bar=True, logger=True)

        return loss

    def get_decoder(self, return_normalization=False):
        """Return a torch model with the decoder and optionally the normalization inverse"""
        if return_normalization:
            if self.norm_in is not None:
                inv_norm = Inverse(module=self.norm_in)
                decoder_model = torch.nn.Sequential(*[self.decoder, inv_norm])
            else:
                raise ValueError(
                    "return_normalization is set to True but self.norm_in is None"
                )
        else:
            decoder_model = self.decoder

        return decoder_model



#!/usr/bin/env python

# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Autocorrelation loss.
"""

__all__ = ["AutocorrelationLoss", "autocorrelation_loss"]


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

from typing import Optional

import torch

from mlcolvar.core.stats.tica import TICA
from mlcolvar.core.loss.eigvals import reduce_eigenvalues_loss


# =============================================================================
# LOSS FUNCTIONS
# =============================================================================


class AutocorrelationLoss(torch.nn.Module):
    """(Weighted) autocorrelation loss.

    Computes the sum (or another reducing functions) of the eigenvalues of the
    autocorrelation matrix. This is the same loss function used in
    :class:`~mlcolvar.cvs.timelagged.deeptica.DeepTICA`.

    """

    def __init__(self, reduce_mode: str = "sum2", invert_sign: bool = True):
        """Constructor.

        Parameters
        ----------
        reduce_mode : str
            This determines how the eigenvalues are reduced, e.g., ``sum``, ``sum2``
            (see also :class:`~mlcolvar.core.loss.eigvals.ReduceEigenvaluesLoss`). The
            default is ``'sum2'``.
        invert_sign: bool, optional
            Whether to return the negative autocorrelation in order to be minimized
            with gradient descent methods. Default is ``True``.
        """
        super().__init__()
        self.reduce_mode = reduce_mode
        self.invert_sign = invert_sign

    def forward(
        self,
        x: torch.Tensor,
        x_lag: torch.Tensor,
        weights: Optional[torch.Tensor] = None,
        weights_lag: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Estimate the autocorrelation.

        Parameters
        ----------
        x : torch.Tensor
            Shape ``(n_batches, n_features)``. The features of the sample at
            time ``t``.
        x_lag : torch.Tensor
            Shape ``(n_batches, n_features)``. The features of the sample at
            time ``t + lag``.
        weights : torch.Tensor, optional
            Shape ``(n_batches,)`` or ``(n_batches, 1)``. The weights associated
            to ``x`` at time ``t``. Default is ``None``.
        weights_lag : torch.Tensor, optional
            Shape ``(n_batches,)`` or ``(n_batches, 1)``. The weights associated
            to ``x`` at time ``t + lag``. Default is ``None``.

        Returns
        -------
        loss : torch.Tensor
            Loss value.
        """
        return autocorrelation_loss(
            x,
            x_lag,
            weights=weights,
            weights_lag=weights_lag,
            reduce_mode=self.reduce_mode,
            invert_sign=self.invert_sign,
        )


def autocorrelation_loss(
    x: torch.Tensor,
    x_lag: torch.Tensor,
    weights: Optional[torch.Tensor] = None,
    weights_lag: Optional[torch.Tensor] = None,
    reduce_mode: str = "sum2",
    invert_sign: bool = True,
) -> torch.Tensor:
    """(Weighted) autocorrelation loss.

    Computes the sum (or another reducing functions) of the eigenvalues of the
    autocorrelation matrix. This is the same loss function used in
    :class:`~mlcolvar.cvs.timelagged.deeptica.DeepTICA`.

    Parameters
    ----------
    x : torch.Tensor
        Shape ``(n_batches, n_features)``. The features of the sample at
        time ``t``.
    x_lag : torch.Tensor
        Shape ``(n_batches, n_features)``. The features of the sample at
        time ``t + lag``.
    weights : torch.Tensor, optional
        Shape ``(n_batches,)`` or ``(n_batches, 1)``. The weights associated
        to ``x`` at time ``t``. Default is ``None``.
    weights_lag : torch.Tensor, optional
        Shape ``(n_batches,)`` or ``(n_batches, 1)``. The weights associated
        to ``x`` at time ``t + lag``. Default is ``None``.
    reduce_mode : str
        This determines how the eigenvalues are reduced, e.g., ``sum``, ``sum2``
        (see also :class:`~mlcolvar.core.loss.eigvals.ReduceEigenvaluesLoss`). The
        default is ``'sum2'``.
    invert_sign: bool, optional
        Whether to return the negative autocorrelation in order to be minimized
        with gradient descent methods. Default is ``True``.

    Returns
    -------
    loss: torch.Tensor
        Loss value.
    """
    tica = TICA(in_features=x.shape[-1])
    eigvals, _ = tica.compute(data=[x, x_lag], weights=[weights, weights_lag])
    loss = reduce_eigenvalues_loss(eigvals, mode=reduce_mode, invert_sign=invert_sign)
    return loss



import mdtraj as md
import itertools


class MDTopologySelector:
    """
    A class to load an MDTraj topology and extract atom pairs (bonds or distances), triplets (angles or arbitrary triples),
    and quads (torsions or arbitrary quadruplets) for a given atom selection.
    """

    def __init__(self, topology):
        """
        Parameters
        ----------
        topology : str | md.Trajectory | md.Topology
            Path to a structure file (e.g., .pdb, .gro), an MDTraj Trajectory, or an MDTraj Topology.
        """

        if isinstance(topology, md.Trajectory):
            self.topology = topology.topology
        elif isinstance(topology, md.Topology):
            self.topology = topology
        elif isinstance(topology, str):
            traj = md.load(topology)
            self.topology = traj.topology
        else:
            raise ValueError("`topology` must be a file path, md.Trajectory, or md.Topology instance.")

        # Precompute bond list as tuples of atom indices
        self.bonds = [(b.atom1.index, b.atom2.index) for b in self.topology.bonds]

    def select(self, selection):
        """
        Select atom indices matching an MDTraj selection string.

        Parameters
        ----------
        selection : str
            MDTraj selection syntax, e.g., "backbone", "name CA", etc.

        Returns
        -------
        numpy.ndarray of int
            Array of atom indices.
        """
        return self.topology.select(selection)

    def get_atom_pairs(self, selection, bonded=True):
        """
        Get atom pairs for a selection.

        Parameters
        ----------
        selection : str
            MDTraj selection syntax.
        bonded : bool, default=True
            If True, return only bonded pairs. If False, return all unique pairs (nonbonded).

        Returns
        -------
        List of tuple(int, int)
            Each tuple is (i, j) of atom indices.
        """
        sel = list(self.select(selection))
        if bonded:
            sel_set = set(sel)
            atom_pairs = [(i, j) for (i, j) in self.bonds if i in sel_set and j in sel_set]
        else:
            atom_pairs = list(itertools.combinations(sel, 2))

        self.n_distances = len(atom_pairs)

        return atom_pairs

    def get_triplets(self, selection, bonded=True):
        """
        Get atom triplets for a selection.

        Parameters
        ----------
        selection : str
            MDTraj selection syntax.
        bonded : bool, default=True
            If True, return triplets that form angles (i-j-k where i-j and j-k are bonds).
            If False, return all unique triplets.

        Returns
        -------
        List of tuple(int, int, int)
            Each tuple is (i, j, k).
        """
        sel = list(self.select(selection))
        if not bonded:
            return list(itertools.combinations(sel, 3))

        sel_set = set(sel)
        # build adjacency dict
        nbrs = {a: set() for a in sel_set}
        for i, j in self.bonds:
            if i in sel_set and j in sel_set:
                nbrs[i].add(j)
                nbrs[j].add(i)

        triplets = set()
        for j in sel_set:
            for i in nbrs[j]:
                for k in nbrs[j]:
                    if i != k:
                        triplets.add((i, j, k))

        self.n_angles = len(list(triplets))

        return list(triplets)

    def get_quads(self, selection, bonded=True):
        """
        Get atom quads for a selection.

        Parameters
        ----------
        selection : str
            MDTraj selection syntax.
        bonded : bool, default=True
            If True, return quads that form torsions (i-j-k-l where i-j, j-k, k-l are bonds).
            If False, return all unique quadruplets.

        Returns
        -------
        List of tuple(int, int, int, int)
            Each tuple is (i, j, k, l).
        """
        sel = list(self.select(selection))
        if not bonded:
            return list(itertools.combinations(sel, 4))

        sel_set = set(sel)
        nbrs = {a: set() for a in sel_set}
        for i, j in self.bonds:
            if i in sel_set and j in sel_set:
                nbrs[i].add(j)
                nbrs[j].add(i)

        quads = set()
        for j in sel_set:
            for k in nbrs[j]:
                for i in nbrs[j]:
                    if i == k:
                        continue
                    for neighbor_l in nbrs[k]:
                        if neighbor_l == j:
                            continue
                        quads.add((i, j, k, neighbor_l))

        self.n_dihedrals = len(quads)

        return list(quads)

    def get_atom_indices(self, selection):
        """
        Get atom indices for a selection.

        Parameters
        ----------
        selection : str
            MDTraj selection syntax.

        Returns
        -------
        List of int
            Atom indices.
        """

        atom_indices = list(self.select(selection))
        self.n_atoms = len(atom_indices)

        return atom_indices

    def topology_indexing(self, config):
        """
        Get the topology indexing for a given configuration.

        Parameters
        ----------
        config : dict
            Configuration dictionary containing selections for cartesian, distances, angles, and dihedrals.

        Returns
        -------
        Dict of indices
            Topology indices.
        """

        self.config = config

        self.topology_idx = {}
        if 'cartesian' in self.config:
            sel = self.config['cartesian']['selection']
            idx = self.get_atom_indices(sel)

            fit_sel = self.config['cartesian'].get('fit_selection', None)
            if fit_sel is not None:
                fit_idx = self.get_atom_indices(fit_sel)
                self.topology_idx['cartesian']['fit_selection'] = fit_idx
                self.topology_idx['cartesian'] = {'selection': sel, 'indices': idx, 'fit_selection': fit_idx}
            else:
                self.topology_idx['cartesian'] = {'selection': sel, 'indices': idx}

        if 'distances' in self.config:
            sel = self.config['distances']['selection']
            bonded = self.config['distances'].get('bonded', False)
            pairs = self.get_atom_pairs(sel, bonded=bonded)
            # pull other args
            cutoff = self.config['distances'].get('cutoff', None)
            periodic = self.config['distances'].get('periodic', False)
            args = {'selection': sel,
                    'pairs': pairs}
            if cutoff is not None:
                args['cutoff'] = cutoff
            args['periodic'] = periodic
            self.topology_idx['distances'] = args

        if 'angles' in self.config:
            sel = self.config['angles']['selection']
            bonded = self.config['angles'].get('bonded', True)
            triplets = self.get_triplets(sel, bonded=bonded)
            periodic = self.config['angles'].get('periodic', False)
            self.topology_idx['angles'] = {
                'selection': sel,
                'triplets': triplets,
                'periodic': periodic
            }

        if 'dihedrals' in self.config:
            sel = self.config['dihedrals']['selection']
            bonded = self.config['dihedrals'].get('bonded', True)
            quads = self.get_quads(sel, bonded=bonded)
            periodic = self.config['dihedrals'].get('periodic', False)
            self.topology_idx['dihedrals'] = {
                'selection': sel,
                'quadruplets': quads,
                'periodic': periodic
            }

        # collect any options
        if 'options' in self.config:
            self.topology_idx['options'] = self.config['options']

        return self.topology_idx



import inspect


def assert_valid_kwargs(target_cls, kwargs, context=""):
    """
    Assert that the keys in kwargs are valid parameters for target_cls.__init__.
    Raises AssertionError if invalid keys are found.

    Args:
        target_cls: class whose __init__ signature to inspect
        kwargs (dict): keyword arguments to validate
        context (str): optional context name for error messages
    """
    sig = inspect.signature(target_cls.__init__)
    params = sig.parameters
    # if **kwargs is accepted, skip strict validation
    if any(p.kind == inspect.Parameter.VAR_KEYWORD for p in params.values()):
        return
    valid_keys = set(params.keys()) - {'self'}
    invalid = set(kwargs.keys()) - valid_keys
    assert not invalid, (
        f"Invalid {context} arguments for {target_cls.__name__}: {invalid}. "
        f"Valid parameters are: {valid_keys}"
    )



#!/usr/bin/env python

# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Committor function Loss Function and Utils.
"""

from mlcolvar.core.transform.descriptors.utils import sanitize_positions_shape
__all__ = ["CommittorLoss", "committor_loss", "SmartDerivatives", "compute_descriptors_derivatives"]

# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

import torch
from typing import Optional

# =============================================================================
# LOSS FUNCTIONS
# =============================================================================


class CommittorLoss(torch.nn.Module):
    """Compute a loss function based on Kolmogorov's variational principle for the determination of the committor function"""

    def __init__(self,
                 mass: torch.Tensor,
                 alpha: float,
                 cell: float = None,
                 gamma: float = 10000,
                 delta_f: float = 0,
                 separate_boundary_dataset: bool = True,
                 descriptors_derivatives: torch.nn.Module = None
                 ):
        """Compute Kolmogorov's variational principle loss and impose boundary conditions on the metastable states

        Parameters
        ----------
        mass : torch.Tensor
            Atomic masses of the atoms in the system
        alpha : float
            Hyperparamer that scales the boundary conditions contribution to loss, i.e. alpha*(loss_bound_A + loss_bound_B)
        cell : float, optional
            CUBIC cell size length, used to scale the positions from reduce coordinates to real coordinates, by default None
        gamma : float, optional
            Hyperparamer that scales the whole loss to avoid too small numbers, i.e. gamma*(loss_var + loss_bound), by default 10000
        delta_f : float, optional
            Delta free energy between A (label 0) and B (label 1), units is kBT, by default 0.
            State B is supposed to be higher in energy.
        separate_boundary_dataset : bool, optional
            Switch to exculde boundary condition labeled data from the variational loss, by default True
        descriptors_derivatives : torch.nn.Module, optional
            `SmartDerivatives` object to save memory and time when using descriptors.
            See also mlcolvar.core.loss.committor_loss.SmartDerivatives

        """
        super().__init__()
        self.register_buffer("mass", mass)
        self.alpha = alpha
        self.cell = cell
        self.gamma = gamma
        self.delta_f = delta_f
        self.descriptors_derivatives = descriptors_derivatives
        self.separate_boundary_dataset = separate_boundary_dataset

    def forward(
        self, x: torch.Tensor, q: torch.Tensor, labels: torch.Tensor, w: torch.Tensor, create_graph: bool = True
    ) -> torch.Tensor:
        return committor_loss(x=x,
                              q=q,
                              labels=labels,
                              w=w,
                              mass=self.mass,
                              alpha=self.alpha,
                              gamma=self.gamma,
                              delta_f=self.delta_f,
                              create_graph=create_graph,
                              cell=self.cell,
                              separate_boundary_dataset=self.separate_boundary_dataset,
                              descriptors_derivatives=self.descriptors_derivatives
                              )


def broadcast(src: torch.Tensor, other: torch.Tensor, dim: int):
    """Broadcast util, from torch_scatter"""
    if dim < 0:
        dim = other.dim() + dim
    if src.dim() == 1:
        for _ in range(0, dim):
            src = src.unsqueeze(0)
    for _ in range(src.dim(), other.dim()):
        src = src.unsqueeze(-1)
    src = src.expand(other.size())
    return src


def scatter_sum(src: torch.Tensor,
                index: torch.Tensor,
                dim: int = -1,
                out: Optional[torch.Tensor] = None,
                dim_size: Optional[int] = None) -> torch.Tensor:
    """Scatter sum function, from torch_scatter module (https://github.com/rusty1s/pytorch_scatter/blob/master/torch_scatter/scatter.py)"""
    index = broadcast(index, src, dim)
    if out is None:
        size = list(src.size())
        if dim_size is not None:
            size[dim] = dim_size
        elif index.numel() == 0:
            size[dim] = 0
        else:
            size[dim] = int(index.max()) + 1
        out = torch.zeros(size, dtype=src.dtype, device=src.device)
        return out.scatter_add_(dim, index, src)
    else:
        return out.scatter_add_(dim, index, src)


def committor_loss(x: torch.Tensor,
                   q: torch.Tensor,
                   labels: torch.Tensor,
                   w: torch.Tensor,
                   mass: torch.Tensor,
                   alpha: float,
                   gamma: float = 10000,
                   delta_f: float = 0,
                   create_graph: bool = True,
                   cell: float = None,
                   separate_boundary_dataset: bool = True,
                   descriptors_derivatives: torch.nn.Module = None
                   ):
    """Compute variational loss for committor optimization with boundary conditions

    Parameters
    ----------
    x : torch.Tensor
        Input of the NN
    q : torch.Tensor
        Committor quess q(x), it is the output of NN
    labels : torch.Tensor
        Labels for states, A and B states for boundary conditions
    w : torch.Tensor
        Reweighing factors to Boltzmann distribution. This should depend on the simulation in which the data were collected.
    mass : torch.Tensor
        List of masses of all the atoms we are using, for each atom we need to repeat three times for x,y,z.
        Can be created using `committor.utils.initialize_committor_masses`
    alpha : float
        Hyperparamer that scales the boundary conditions contribution to loss, i.e. alpha*(loss_bound_A + loss_bound_B)
    gamma : float
        Hyperparamer that scales the whole loss to avoid too small numbers, i.e. gamma*(loss_var + loss_bound)
        By default 10000
    delta_f : float
        Delta free energy between A (label 0) and B (label 1), units is kBT, by default 0.
    create_graph : bool
        Make loss backwardable, deactivate for validation to save memory, default True
    cell : float
        CUBIC cell size length, used to scale the positions from reduce coordinates to real coordinates, default None
    separate_boundary_dataset : bool, optional
            Switch to exculde boundary condition labeled data from the variational loss, by default True
    descriptors_derivatives : torch.nn.Module, optional
        `SmartDerivatives` object to save memory and time when using descriptors.
        See also mlcolvar.core.loss.committor_loss.SmartDerivatives

    Returns
    -------
    loss : torch.Tensor
        Loss value.
    gamma*loss_var : torch.Tensor
        The variational loss term
    gamma*alpha*loss_A : torch.Tensor
        The boundary loss term on basin A
    gamma*alpha*loss_B : torch.Tensor
        The boundary loss term on basin B
    """
    # inherit right device
    device = x.device

    mass = mass.to(device)

    # Create masks to access different states data
    mask_A = torch.nonzero(labels.squeeze() == 0, as_tuple=True)
    mask_B = torch.nonzero(labels.squeeze() == 1, as_tuple=True)
    if separate_boundary_dataset:
        mask_var = torch.nonzero(labels.squeeze() > 1, as_tuple=True)
    else:
        mask_var = torch.ones(len(x), dtype=torch.bool)

    # Update weights of basin B using the information on the delta_f
    delta_f = torch.Tensor([delta_f])
    if delta_f < 0:  # B higher in energy --> A-B < 0
        w[mask_B] = w[mask_B] * torch.exp(delta_f.to(device))
    elif delta_f > 0:  # A higher in energy --> A-B > 0
        w[mask_A] = w[mask_A] * torch.exp(-delta_f.to(device))

    # VARIATIONAL PRINICIPLE LOSS
    # Each loss contribution is scaled by the number of samples

    # We need the gradient of q(x)
    grad_outputs = torch.ones_like(q[mask_var])
    grad = torch.autograd.grad(q[mask_var], x, grad_outputs=grad_outputs, retain_graph=True, create_graph=create_graph)[0]
    grad = grad[mask_var]

    # TODO this fixes cell size issue
    if cell is not None:
        grad = grad / cell

    if descriptors_derivatives is not None:
        grad_square = descriptors_derivatives(grad)
    else:
        # we get the square of grad(q) and we multiply by the weight
        grad_square = torch.pow(grad, 2)

    # we sanitize the shapes of mass and weights tensors
    # mass should have size [1, n_atoms*spatial_dims]
    mass = mass.unsqueeze(0)
    # weights should have size [n_batch, 1]
    w = w.unsqueeze(-1)

    grad_square = torch.sum((grad_square * (1 / mass)), axis=1, keepdim=True)
    grad_square = grad_square * w[mask_var]

    # variational contribution to loss: we sum over the batch
    loss_var = torch.mean(grad_square)

    # boundary conditions
    q_A = q[mask_A]
    q_B = q[mask_B]
    loss_A = torch.mean(torch.pow(q_A, 2))
    loss_B = torch.mean(torch.pow((q_B - 1), 2))

    loss = gamma * (loss_var + alpha * (loss_A + loss_B))

    # TODO maybe there is no need to detach them for logging
    return loss, gamma * loss_var.detach(), alpha * gamma * loss_A.detach(), alpha * gamma * loss_B.detach()


class SmartDerivatives(torch.nn.Module):
    """
    Utils to compute efficently (time and memory wise) the derivatives of q wrt some input descriptors.
    Rather than computing explicitly the derivatives wrt the positions, we compute those wrt the descriptors (right input)
    and multiply them by the matrix of the derivatives of the descriptors wrt the positions (left input).
    """

    def __init__(self,
                 der_desc_wrt_pos: torch.Tensor,
                 n_atoms: int,
                 setup_device: str = 'cpu'
                 ):
        """Initialize the fixed matrices for smart derivatives, i.e. matrix of derivatives of descriptors wrt positions.
        The derivatives wrt positions are recovered by multiplying the derivatives of q wrt the descriptors (right input, computed at each epoch)
        by the non-zero elements of the derivatives of the descriptors wrt the positions (left input, compute once at the beginning on the whole dataset).
        The multiplication are done using scatte functions and keepoing track of the indeces of the batches, descriptors, atoms and dimensions.

        NB. It should be used with only training set and single batch with shuffle and random_split disabled.

        Parameters
        ----------
        der_desc_wrt_pos : torch.Tensor
            Tensor containing the derivatives of the descriptors wrt the atomic positions
        n_atoms : int
            Number of atoms in the systems, all the atoms should be used in at least one of the descriptors
        setup_device : str
            Device on which to perform the expensive calculations. Either 'cpu' or 'cuda', by default 'cpu'
        """
        super().__init__()
        self.batch_size = len(der_desc_wrt_pos)
        self.n_atoms = n_atoms

        # setup the fixed part of the computation, i.e. left input and indeces for the scatter
        self.left, self.mat_ind, self.scatter_indeces = self._setup_left(der_desc_wrt_pos, setup_device=setup_device)

    def _setup_left(self, left_input: torch.Tensor, setup_device: str = 'cpu'):
        """Setup the fixed part of the computation, i.e. left input"""
        # all the setup should be done on the CPU by defualt
        left_input = left_input.to(torch.device(setup_device))

        # the indeces in mat_ind are: batch, atom, descriptor and dimension
        left, mat_ind = self._create_nonzero_left(left_input)

        # it is possible that some atoms are not involved in anything
        n_effective_atoms = len(torch.unique(mat_ind[1]))
        if n_effective_atoms < self.n_atoms:
            raise ValueError(f"Some of the input atoms are useless LOL. The not used atom IDs are : {[i for i in range(self.n_atoms) if i not in torch.unique(mat_ind[1]).numpy()]}  ")

        scatter_indeces = self._get_scatter_indices(batch_ind=mat_ind[0], atom_ind=mat_ind[1], dim_ind=mat_ind[3])
        return left, mat_ind, scatter_indeces

    def _create_nonzero_left(self, x):
        """Find the indeces of the non-zero elements of the left input
        """
        # find indeces of nonzero entries of d_dist_d_x
        mat_ind = x.nonzero(as_tuple=True)

        # flatten matrix --> big nonzero vector
        x = x.ravel()
        # find indeces of nonzero entries of the flattened matrix
        vec_ind = x.nonzero(as_tuple=True)

        # create vector with the nonzero entries only
        x_vec = x[vec_ind[0].long()]

        # del(vec_ind)
        return x_vec, mat_ind

    def _get_scatter_indices(self, batch_ind, atom_ind, dim_ind):
        """Compute the general indices to map the long vector of nonzero derivatives to the right atom, dimension and descriptor also in the case of non homogenoeus input.
        We need to gather the derivatives with respect to the same atom coming from different descriptors to obtain the total gradient.
        """
        # ====================================== INITIAL NOTE  ======================================
        # in the comment there's the example of the distances in a 3 atoms system with 4 batches
        # i.e. 3desc*3*atom*3dim*2pairs*4batch = 72 values needs to be mappped to 3atoms*3dims*4batch = 36

        # Ref_idx:  tensor([ 0,  1,  2,  0,  1,  2,  3,  4,  5,  3,  4,  5,  6,  7,  8,  6,  7,  8,
        #          9, 10, 11,  9, 10, 11, 12, 13, 14, 12, 13, 14, 15, 16, 17, 15, 16, 17,
        #         18, 19, 20, 18, 19, 20, 21, 22, 23, 21, 22, 23, 24, 25, 26, 24, 25, 26,
        #         27, 28, 29, 27, 28, 29, 30, 31, 32, 30, 31, 32, 33, 34, 35, 33, 34, 35])
        # ==========================================================================================

        # these would be the indeces in the case of uniform batches and number of atom/descriptor dependence
        # it just repeats the atom index in a cycle
        # e.g. [0, 1, 2, 0, 1, 2, 3, 4, 5, 3, 4, 5, 6, 7, 8, 6, 7, 8, 0, 1, 2, 0, 1, 2,
        #       3, 4, 5, 3, 4, 5, 6, 7, 8, 6, 7, 8, 0, 1, 2, 0, 1, 2, 3, 4, 5, 3, 4, 5,
        #       6, 7, 8, 6, 7, 8, 0, 1, 2, 0, 1, 2, 3, 4, 5, 3, 4, 5, 6, 7, 8, 6, 7, 8]
        not_shifted_indeces = atom_ind * 3 + dim_ind

        # get the number of elements in each batch
        # e.g. [17, 18, 18, 18]
        batch_elements = scatter_sum(torch.ones_like(batch_ind), batch_ind)
        batch_elements[0] -= 1  # to make the later indexing consistent

        # compute the pointer idxs to the beginning of each batch by summing the number of elements in each batch
        # e.g. [ 0., 17., 35., 53.] NB. These are indeces!
        batch_pointers = torch.Tensor([batch_elements[:i].sum() for i in range(len(batch_elements))])
        del (batch_elements)

        # number of entries in the scattered vector before each batch
        # e.g. [ 0.,  9., 18., 27.]
        markers = not_shifted_indeces[batch_pointers.long()]  # highest not_shifted index for each batch
        del (not_shifted_indeces)
        del (batch_pointers)
        cumulative_markers = torch.Tensor([markers[:i + 1].sum() for i in range(len(markers))]).to(batch_ind.device)  # stupid sum of indeces
        del (markers)
        cumulative_markers += torch.unique(batch_ind)  # markers correction by the number of batches

        # get the index shift in the scattered vector based on the batch
        # e.g. [ 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  9,  9,  9,  9,  9,  9,
        #        9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18,
        #       18, 18, 18, 18, 18, 18, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27 ]
        batch_shift = torch.gather(cumulative_markers, 0, batch_ind)
        del (cumulative_markers)

        # finally compute the scatter indeces by including also their shift due to the batch
        shifted_indeces = atom_ind * 3 + dim_ind + batch_shift

        return shifted_indeces

    def forward(self, x: torch.Tensor):
        # ensure device consistency
        left = self.left.to(x.device)

        # get the vector with the derivatives of q wrt the descriptors
        right = self._create_right(x=x, batch_ind=self.mat_ind[0], des_ind=self.mat_ind[2])

        # do element-wise product
        src = left * right

        # compute square modulus
        out = self._compute_square_modulus(x=src, indeces=self.scatter_indeces, n_atoms=self.n_atoms, batch_size=self.batch_size)

        return out

    def _create_right(self, x: torch.Tensor, batch_ind: torch.Tensor, des_ind: torch.Tensor):
        # keep only the non zero elements of right input
        desc_vec = x[batch_ind, des_ind]
        return desc_vec

    def _compute_square_modulus(self, x: torch.Tensor, indeces: torch.Tensor, n_atoms: int, batch_size: torch.Tensor):
        indeces = indeces.long().to(x.device)

        # this sums the elements of x according to the indeces, this way we get the contributions of different descriptors to the same atom
        out = scatter_sum(x, indeces.long())
        # now make the square
        out = out.pow(2)
        # reshape, this needs to have the correct number of atoms as we need to mulply it by the mass vector later
        out = out.reshape((batch_size, n_atoms * 3))
        return out


def compute_descriptors_derivatives(dataset, descriptor_function, n_atoms, separate_boundary_dataset=True):
    """Compute the derivatives of a set of descriptors wrt input positions in a dataset for committor optimization

    Parameters
    ----------
    dataset :
        DictDataset with the positions under the 'data' key
    descriptor_function : torch.nn.Module
        Transform module for the computation of the descriptors
    n_atoms : int
        Number of atoms in the system
    separate_boundary_dataset : bool, optional
            Switch to exculde boundary condition labeled data from the variational loss, by default True

    Returns
    -------
    desc : torch.Tensor
        Computed descriptors
    d_desc_d_pos : torch.Tensor
        Derivatives of desc wrt to pos
    """
    pos = dataset['data']
    labels = dataset['labels']
    pos = sanitize_positions_shape(pos=pos, n_atoms=n_atoms)[0]
    pos.requires_grad = True

    desc = descriptor_function(pos)
    if separate_boundary_dataset:
        mask_var = torch.nonzero(labels.squeeze() > 1, as_tuple=True)[0]
        der_desc = desc[mask_var]
        if len(der_desc) == 0:
            raise (ValueError('No points left after separating boundary and variational datasets. \n If you are using only unbiased data set separate_boundary_dataset=False here and in Committor or don\'t use SmartDerivatives!!'))
    else:
        der_desc = desc

    # compute derivatives of descriptors wrt positions, loop over the number of decriptors
    aux = []
    for i in range(len(der_desc[0])):
        aux_der = torch.autograd.grad(der_desc[:, i], pos, grad_outputs=torch.ones_like(der_desc[:, i]), retain_graph=True)[0]
        if separate_boundary_dataset:
            aux_der = aux_der[mask_var]
        aux.append(aux_der)

    d_desc_d_pos = torch.stack(aux, axis=2)
    return pos, desc, d_desc_d_pos.squeeze(-1)


def test_smart_derivatives():
    from mlcolvar.core.transform import PairwiseDistances
    from mlcolvar.core.nn import FeedForward
    from mlcolvar.data import DictDataset

    # compute some descriptors from positions --> distances
    n_atoms = 10
    pos = torch.Tensor([[1.4970, 1.3861, -0.0273, -1.4933, 1.5070, -0.1133, -1.4473, -1.4193,
                        -0.0553, 1.4940, 1.4990, -0.2403, 1.4780, -1.4173, -0.3363, -1.4243,
                        -1.4093, -0.4293, 1.3530, -1.4313, -0.4183, 1.3060, 1.4750, -0.4333,
                        1.2970, -1.3233, -0.4643, 1.1670, -1.3253, -0.5354]])
    pos = pos.repeat(4, 1)
    labels = torch.arange(0, 4)

    dataset = DictDataset({'data': pos, 'labels': labels})

    cell = torch.Tensor([3.0233])
    ref_distances = torch.Tensor([[0.1521, 0.2335, 0.2412, 0.3798, 0.4733, 0.4649, 0.4575, 0.5741, 0.6815,
                                   0.1220, 0.1323, 0.2495, 0.3407, 0.3627, 0.3919, 0.4634, 0.5885, 0.2280,
                                   0.2976, 0.3748, 0.4262, 0.4821, 0.5043, 0.6376, 0.1447, 0.2449, 0.2454,
                                   0.2705, 0.3597, 0.4833, 0.1528, 0.1502, 0.2370, 0.2408, 0.3805, 0.2472,
                                   0.3243, 0.3159, 0.4527, 0.1270, 0.1301, 0.2440, 0.2273, 0.2819, 0.1482]])
    ref_distances = ref_distances.repeat(4, 1)

    ComputeDescriptors = PairwiseDistances(n_atoms=n_atoms,
                                           PBC=True,
                                           cell=cell,
                                           scaled_coords=False)

    for separate_boundary_dataset in [False, True]:
        if separate_boundary_dataset:
            mask = [labels > 1]
        else:
            mask = torch.ones_like(labels, dtype=torch.bool)

        pos, desc, d_desc_d_x = compute_descriptors_derivatives(dataset=dataset,
                                                                descriptor_function=ComputeDescriptors,
                                                                n_atoms=n_atoms,
                                                                separate_boundary_dataset=separate_boundary_dataset)

        assert (torch.allclose(desc, ref_distances, atol=1e-3))

        # apply simple NN
        NN = FeedForward(layers=[45, 2, 1])
        out = NN(desc)

        # compute derivatives of out wrt input
        d_out_d_x = torch.autograd.grad(out, pos, grad_outputs=torch.ones_like(out), retain_graph=True, create_graph=True)[0]
        # compute derivatives of out wrt descriptors
        d_out_d_d = torch.autograd.grad(out, desc, grad_outputs=torch.ones_like(out), retain_graph=True, create_graph=True)[0]
        ref = torch.einsum('badx,bd->bax ', d_desc_d_x, d_out_d_d[mask])
        ref = ref.pow(2).sum(dim=(-2, -1))

        Ref = d_out_d_x[mask].pow(2).sum(dim=(-2, -1))

        # apply smart derivatives
        smart_derivatives = SmartDerivatives(d_desc_d_x, n_atoms=n_atoms)
        right_input = d_out_d_d.squeeze(-1)
        smart_out = smart_derivatives(right_input).sum(dim=1)

        # do checks
        assert (torch.allclose(smart_out, ref))
        assert (torch.allclose(smart_out, Ref))

        smart_out.sum().backward()


if __name__ == "__main__":
    test_smart_derivatives()



"""Common functions for package biobb_pytorch.models"""

from pathlib import Path
from typing import Callable, Optional, Union

import numpy as np
import torch


def ndarray_normalization(
    ndarray: np.ndarray, max_values: np.ndarray, min_values: np.ndarray
) -> np.ndarray:
    """
    Normalize an ndarray along a specified axis.

    Args:
        ndarray (np.ndarray): The input ndarray to be normalized.
        max_values (np.ndarray): The maximum values for normalization.
        min_values (np.ndarray): The minimum values for normalization.

    Returns:
        np.ndarray: The normalized ndarray.
    """
    return (ndarray - min_values) / (max_values - min_values)


def ndarray_denormalization(
    normalized_ndarray: np.ndarray, max_values: np.ndarray, min_values: np.ndarray
) -> np.ndarray:
    """
    Denormalizes a normalized ndarray using the given max and min values.

    Args:
        normalized_ndarray (np.ndarray): The normalized ndarray to be denormalized.
        max_values (np.ndarray): The maximum value used for normalization.
        min_values (np.ndarray): The minimum value used for normalization.

    Returns:
        np.ndarray: The denormalized ndarray.
    """
    return normalized_ndarray * (max_values - min_values) + min_values


def get_loss_function(loss_function: str) -> Callable:
    """
    Get the loss function from the given string.

    Args:
        loss_function (str): The loss function to be used.

    Returns:
        Callable: The loss function.
    """
    loss_function_dict = dict(
        filter(lambda pair: pair[0].endswith("Loss"), vars(torch.nn).items())
    )
    try:
        return loss_function_dict[loss_function]
    except KeyError:
        raise ValueError(f"Invalid loss function: {loss_function}")


def get_optimizer_function(optimizer_function: str) -> Callable:
    """
    Get the optimizer function from the given string.

    Args:
        optimizer_function (str): The optimizer function to be used.

    Returns:
        Callable: The optimizer function.
    """
    optimizer_function_dict = dict(
        filter(lambda pair: not pair[0].startswith("_"), vars(torch.optim).items())
    )
    try:
        return optimizer_function_dict[optimizer_function]
    except KeyError:
        raise ValueError(f"Invalid optimizer function: {optimizer_function}")


def execute_model(
    model: torch.nn.Module,
    dataloader: torch.utils.data.DataLoader,
    input_dimensions: int,
    latent_dimensions: int,
    loss_function: Optional[torch.nn.modules.loss._Loss] = None,
) -> tuple[float, np.ndarray, np.ndarray]:
    model.eval()
    losses: list[float] = []
    z_list: list[float] = []
    x_hat_list: list[float] = []
    with torch.no_grad():
        for data in dataloader:
            data = data[0].to(model.device)
            latent, output = model(data)
            if loss_function:
                loss = loss_function(output, data)
                losses.append(loss.item())
            z_list.append(latent.cpu().numpy())
            x_hat_list.append(output.cpu().numpy())
    loss = float(np.mean(losses)) if losses else -1.0
    latent_space: np.ndarray = np.reshape(
        np.concatenate(z_list, axis=0), (-1, latent_dimensions)
    )
    reconstructed_data: np.ndarray = np.reshape(
        np.concatenate(x_hat_list, axis=0), (-1, input_dimensions)
    )
    return loss, latent_space, reconstructed_data


def format_time(seconds: Union[float, int]) -> str:
    """Converts time in seconds to a string of the format 'HH:MM:SS'."""
    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    if hours:
        return "{:02}h {:02}m {:02}s".format(int(hours), int(minutes), int(seconds))
    elif minutes:
        return "{:02}m {:02}s".format(int(minutes), int(seconds))
    else:
        return "{:02}s".format(int(seconds))


def human_readable_file_size(file_path: Union[str, Path]) -> str:
    """Get the size of a file and return it in a human-readable format."""
    file_path = Path(file_path)  # Ensure file_path is a Path object
    size_in_bytes: float = file_path.stat().st_size
    units = ["Bytes", "KB", "MB", "GB", "PB"]
    for unit in units:
        if size_in_bytes < 1024:
            return f"{size_in_bytes:.2f} {unit}"
        size_in_bytes /= 1024
    return f"{size_in_bytes:.2f} {unit}"



import torch
import torch.nn.functional as F
from typing import Union


class Shifted_Softplus(torch.nn.Softplus):
    """Element-wise softplus function shifted as to pass from the origin."""

    def __init__(self, beta=1, threshold=20):
        super(Shifted_Softplus, self).__init__(beta, threshold)

    def forward(self, input):
        sp0 = F.softplus(torch.zeros(1), self.beta, self.threshold).item()
        return F.softplus(input, self.beta, self.threshold) - sp0


class Custom_Sigmoid(torch.nn.Module):
    def __init__(self, p=3):
        super(Custom_Sigmoid, self).__init__()
        self.p = p

    def forward(self, input):
        return 1 / (1 + torch.exp(-self.p * (input)))


def get_activation(activation: str):
    """Return activation module given string."""
    activ = None
    if activation == "relu":
        activ = torch.nn.ReLU(True)
    elif activation == "elu":
        activ = torch.nn.ELU(True)
    elif activation == "tanh":
        activ = torch.nn.Tanh()
    elif activation == "softplus":
        activ = torch.nn.Softplus()
    elif activation == "shifted_softplus":
        activ = Shifted_Softplus()
    elif activation == "sigmoid":
        activ = torch.nn.Sigmoid()
    elif activation == "logsoftmax":
        activ = torch.nn.LogSoftmax(dim=1)
    elif activation == "linear":
        print("WARNING: no activation selected")
    elif activation is None:
        pass
    else:
        raise ValueError(
            f"Unknown activation: {activation}. options: 'relu','elu','tanh','softplus','shifted_softplus','logsoftmax','linear'. "
        )
    return activ


def parse_nn_options(options: Union[str, list], n_layers: int, last_layer_activation: Union[bool, str]):
    """Parse args per layer of the NN.

    If a single value is given, repeat options to all layers but for the output one,
    unless ``last_layer_activation is True``, in which case the option is repeated
    also for the output layer.
    """
    if last_layer_activation is False:
        last_layer_activation = None

    # If an iterable is given cheeck that its length matches the number of NN layers
    if hasattr(options, "__iter__") and not isinstance(options, str):
        if len(options) != n_layers:
            raise ValueError(
                f"Length of options: {options} ({len(options)} should be equal to number of layers ({n_layers}))."
            )
        options_list = options

    # if a single value is given, repeat options to all layers but for the output one
    else:
        if last_layer_activation:
            if isinstance(last_layer_activation, str):
                options_list = [options for _ in range(n_layers - 1)]
                options_list.append(last_layer_activation)
            else:
                options_list = [options for _ in range(n_layers)]
        else:
            options_list = [options for _ in range(n_layers - 1)]
            options_list.append(None)

    return options_list



import mdtraj as md
import numpy as np
import torch
from biobb_pytorch.mdae.featurization.normalization import Normalization
from mlcolvar.core.transform.utils import Statistics


class Featurizer:
    """
    A class to extract geometric features (distances, angles, dihedrals) from MD trajectories using MDTraj.
    Supports selections by atom indices or MDTraj selection strings.
    Parameters:
    -------------
    trajectory_file : str
        Path to the trajectory file (e.g., .dcd, .xtc).
    topology_file : str
        Path to the topology file (e.g., .pdb, .gro).
    input_labels_npy_path : str, optional
        Path to a .npy file containing labels for each frame.
    input_weights_npy_path : str, optional
        Path to a .npy file containing weights for each frame.
    """

    def __init__(self, trajectory_file, topology_file, input_labels_npy_path=None, input_weights_npy_path=None):
        """
        Initialize with an MDTraj Trajectory object.

        Parameters:
        -------------
        trajectory_file : str
            Path to the trajectory file (e.g., .dcd, .xtc).
        topology_file : str
            Path to the topology file (e.g., .pdb, .gro).
        """

        # Load trajectory and topology
        trajectory = md.load(trajectory_file,
                             top=topology_file)

        self.trajectory = trajectory
        self.topology = trajectory.topology
        self.input_labels_npy_path = input_labels_npy_path
        self.input_weights_npy_path = input_weights_npy_path

        self.complete_top = md.Trajectory(xyz=trajectory.xyz[0], topology=trajectory.topology)

    def select_atoms(self, selection):
        """
        Convert a selection specifier into atom indices.

        Parameters:
        -------------
        selection : str or list[int] or np.ndarray
            If str: MDTraj topology query (e.g., 'name CA', 'resid 0 to 10').
            If list or array: explicit atom indices.

        Returns:
        --------
        np.ndarray
            Array of selected atom indices.
        """
        if isinstance(selection, str):
            idx = self.topology.select(selection)
        else:
            idx = np.array(selection, dtype=int)
        return idx

    def filter_topology(self, selection, topology):
        """
        Filter the topology based on a selection.

        Parameters:
        -------------
        selection : str
            MDTraj topology query (e.g., 'name CA', 'resid 0 to 10').
        topology : md.Topology
            MDTraj topology object.

        Returns:
        --------
        md.Topology
            Filtered topology.
        """
        idx = self.select_atoms(selection)
        return topology.atom_slice(idx)

    def _dicts_to_tuples(self, dict_list, expected_length):
        """
        Internal helper: convert list of dicts mapping resid->atom to tuple of atom indices.

        Each dict must have exactly expected_length entries. Keys are residue indices,
        values are atom names. Residue indices must match MDTraj topology resid numbering.
        """
        tuples = []
        for d in dict_list:
            if len(d) != expected_length:
                raise ValueError(f"Expected dict with {expected_length} entries, got {len(d)}")
            items = list(d.items())
            idxs = []
            for resid, atom_name in items:
                sel = f"resid {resid} and name {atom_name}"
                arr = self.select_atoms(sel)
                if len(arr) == 0:
                    raise ValueError(f"No atom found for {sel}")
                # take first match
                idxs.append(int(arr[0]))
            tuples.append(tuple(idxs))
        return tuples

    def idx_distances(self, pairs):
        """
        Convert pairs of atom indices to MDTraj topology indices.

        Parameters:
        -------------
        pairs : list of tuples or dicts
            Each tuple or dict must have 2 entries (resid, atom_name).

        Returns:
        --------
        np.ndarray
            Array of shape (n_pairs, 2) with atom indices.
        """
        if len(pairs) > 0 and isinstance(pairs[0], dict):
            pairs = self._dicts_to_tuples(pairs, 2)

        idx_pairs = []
        for a, b in pairs:
            ia = int(self.select_atoms(a)) if isinstance(a, str) else int(a)
            ib = int(self.select_atoms(b)) if isinstance(b, str) else int(b)
            idx_pairs.append((ia, ib))
        idx_pairs = np.array(idx_pairs)
        return idx_pairs

    def compute_distances(self, idx_pairs, cutoff, periodic: bool = True):
        """
        Compute inter-atomic distances for given pairs.

        Accepts pairs as list of 2-tuples or list of dicts with 2 entries {resid: atom_name}.
        """
        distances = md.compute_distances(self.trajectory, idx_pairs, periodic=periodic)

        # apply cutoff and get only pairs within cutoff
        if cutoff is not None:
            keep_cols = np.any(distances < cutoff, axis=0)
            idx_pairs = idx_pairs[keep_cols]
            distances = distances[:, keep_cols]

        return distances, idx_pairs

    def polar2cartesian(self, a):
        """
        Convert polar coordinates to Cartesian coordinates.

        Parameters:
        -------------
        a : np.ndarray
            Array of shape (n_frames, n_angles) representing polar coordinates.
            Each row corresponds to a frame, each column to an angle.

        Returns:
        --------
        np.ndarray
            Array of shape (n_frames, n_angles * 2) representing Cartesian coordinates.
            Each row corresponds to a frame, each column to sin and cos a.
        """
        # Convert angles to radians
        a = np.deg2rad(a)
        # Compute sin and cos
        x = np.sin(a)
        y = np.cos(a)
        # Stack sin and cos values
        cart_angles = np.column_stack((x, y))
        return cart_angles

    def cartesian2polar(self, cart_angles):
        """
        Convert Cartesian coordinates to polar coordinates.

        Parameters:
        -------------
        cart_angles : np.ndarray
            Array of shape (n_frames, n_angles * 2) representing Cartesian coordinates.
            Each row corresponds to a frame, each column to sin and cos a.

        Returns:
        --------
        np.ndarray
            Array of shape (n_frames, n_angles) representing polar coordinates.
            Each row corresponds to a frame, each column to an angle.
        """
        # Compute angles from sin and cos
        angles = np.arctan2(cart_angles[:, 0], cart_angles[:, 1])
        # Convert angles to degrees
        angles = np.rad2deg(angles)
        return angles

    def idx_angles(self, triplets):
        """
        Convert triplets of atom indices to MDTraj topology indices.

        Parameters:
        -------------
        triplets : list of tuples or dicts
            Each tuple or dict must have 3 entries (resid, atom_name).

        Returns:
        --------
        np.ndarray
            Array of shape (n_triplets, 3) with atom indices.
        """
        if len(triplets) > 0 and isinstance(triplets[0], dict):
            triplets = self._dicts_to_tuples(triplets, 3)

        idx_triplets = []
        for i, j, k in triplets:
            ii = int(self.select_atoms(i)) if isinstance(i, str) else int(i)
            jj = int(self.select_atoms(j)) if isinstance(j, str) else int(j)
            kk = int(self.select_atoms(k)) if isinstance(k, str) else int(k)
            idx_triplets.append((ii, jj, kk))
        idx_triplets = np.array(idx_triplets)
        return idx_triplets

    def compute_angles(self, idx_triplets, periodic: bool = True):
        """
        Compute angles between triplets of atoms.

        Accepts triplets as list of 3-tuples or list of dicts with 3 entries {resid: atom_name}.
        """

        return md.compute_angles(self.trajectory, idx_triplets, periodic=periodic)

    def idx_dihedrals(self, quadruplets):
        """
        Convert quadruplets of atom indices to MDTraj topology indices.

        Parameters:
        -------------
        quadruplets : list of tuples or dicts
            Each tuple or dict must have 4 entries (resid, atom_name).

        Returns:
        --------
        np.ndarray
            Array of shape (n_quadruplets, 4) with atom indices.
        """
        if len(quadruplets) > 0 and isinstance(quadruplets[0], dict):
            quadruplets = self._dicts_to_tuples(quadruplets, 4)

        idx_quads = []
        for i, j, k, l in quadruplets:
            ii = int(self.select_atoms(i)) if isinstance(i, str) else int(i)
            jj = int(self.select_atoms(j)) if isinstance(j, str) else int(j)
            kk = int(self.select_atoms(k)) if isinstance(k, str) else int(k)
            ll = int(self.select_atoms(l)) if isinstance(l, str) else int(l)
            idx_quads.append((ii, jj, kk, ll))
        idx_quads = np.array(idx_quads)
        return idx_quads

    def compute_dihedrals(self, idx_quads, periodic: bool = True):
        """
        Compute dihedral angles for quadruplets of atoms.

        Accepts quads as list of 4-tuples or list of dicts with 4 entries {resid: atom_name}.
        """

        return md.compute_dihedrals(self.trajectory, idx_quads, periodic=periodic)

    def compute_cartesian(self, indices):
        """
        Compute Cartesian coordinates for selected atoms.

        Parameters:
        -------------
        indices : list[int]
            List of atom indices to compute Cartesian coordinates for.

        Returns:
        --------
        np.ndarray
            Cartesian coordinates of the selected atoms.
        """
        return self.trajectory.xyz[:, indices, :]

    def combine_features(self, *feature_arrays):
        """
        Concatenate multiple feature arrays along the feature axis.
        """
        return np.concatenate(feature_arrays, axis=1)

    def timelag(self, data: np.ndarray, lag: int):
        """
        Split into X and Y where Y[t] = X[t+lag].

        Parameters
        ----------
        data : np.ndarray, shape (n_times, n_features)
        lag : int

        Returns
        -------
        X : np.ndarray, shape (n_times-lag, n_features)
        Y : np.ndarray, shape (n_times-lag, n_features)
        """
        if lag < 1 or lag >= data.shape[0]:
            raise ValueError("lag must be between 1 and n_times-1")

        X = data[:-lag]
        Y = data[lag:]
        return X, Y

    def get_n_features(self):
        """
        Get the number of features in the combined feature array.
        """
        return self.n_features

    def get_n_frames(self):
        """
        Get the number of frames in the combined feature array.
        """
        return self.n_frames

    def get_atom_info(self, selection):
        """
        Get the atom information from the topology.
        """
        idx = self.select_atoms(selection)
        top = self.trajectory.atom_slice(idx)
        atom_info = []
        for i in top.topology.atoms:
            atom_info.append([i.name, i.residue.name, i.residue.index + 1])
        return np.array(atom_info, dtype=object)

    def set_statistics(self, combined: np.ndarray, feature_dict: dict):
        """
        Set statistics for the combined feature array.
        """
        stats = Statistics(torch.FloatTensor(combined)).to_dict()
        stats['shape'] = [self.n_frames, self.n_features]
        stats['selection'] = feature_dict['cartesian']['selection'] if 'cartesian' in feature_dict else None
        stats['topology'] = self.filter_topology(stats.get('selection', "name CA"), self.complete_top)
        stats["parametric"] = [torch.mean(torch.from_numpy(combined.flatten())), torch.std(torch.from_numpy(combined.flatten()))]

        if self.idx_cartesian is not None:
            stats['cartesian_indices'] = self.idx_cartesian
        if self.idx_dist is not None:
            stats['distance_indices'] = self.idx_dist.tolist()
        if self.idx_triplets is not None:
            stats['angle_indices'] = self.idx_triplets.tolist()
        if self.idx_quads is not None:
            stats['dihedral_indices'] = self.idx_quads.tolist()

        return stats

    def compute_features(self, feature_dict: dict):
        """
        Compute and combine multiple feature types in one call.

        feature_dict keys:
          - 'distances': list of tuple or dict (2 entries)
          - 'angles':    list of tuple or dict (3 entries)
          - 'dihedrals': list of tuple or dict (4 entries)

        Returns:
          combined : np.ndarray shape=(n_frames, total_features)
          features  : dict mapping feature type to its array
        """
        self.idx_cartesian = None
        self.idx_dist = None
        self.idx_triplets = None
        self.idx_quads = None

        self.features = {}
        arrays = []

        if 'distances' in feature_dict:
            self.idx_dist = self.idx_distances(feature_dict['distances']['pairs'])
            d, self.idx_dist = self.compute_distances(self.idx_dist,
                                                      cutoff=feature_dict['distances']['cutoff'],
                                                      periodic=feature_dict['distances']['periodic'])
            self.n_distances = d.shape[1]
            self.features['distances'] = d
            arrays.append(d)

        if 'angles' in feature_dict:
            self.idx_triplets = self.idx_angles(feature_dict['angles']['triplets'])
            a = self.compute_angles(self.idx_triplets,
                                    periodic=feature_dict['angles']['periodic'])
            self.n_angles = a.shape[1]
            a = self.polar2cartesian(a)
            self.features['angles'] = a
            arrays.append(a)

        if 'dihedrals' in feature_dict:
            self.idx_quads = self.idx_dihedrals(feature_dict['dihedrals']['quadruplets'])
            phi = self.compute_dihedrals(self.idx_quads,
                                         periodic=feature_dict['dihedrals']['periodic'])
            self.n_dihedrals = phi.shape[1]
            phi = self.polar2cartesian(phi)
            self.features['dihedrals'] = phi
            arrays.append(phi)

        if 'cartesian' in feature_dict:
            self.idx_cartesian = feature_dict['cartesian']['indices']
            cart = self.compute_cartesian(self.idx_cartesian)
            self.features['cartesian'] = cart.reshape(self.trajectory.n_frames, -1)
            self.n_cartesian = cart.shape[1]
            arrays.append(self.features['cartesian'])

        combined = self.combine_features(*arrays)

        self.n_features = combined.shape[1]
        self.n_frames = combined.shape[0]

        stats = self.set_statistics(combined, feature_dict)

        if self.input_labels_npy_path:
            labels = np.load(self.input_labels_npy_path)

        if self.input_weights_npy_path:
            weights = np.load(self.input_weights_npy_path)

        if 'norm_in' in feature_dict.get('options', {}):

            if feature_dict['options']['norm_in']['mode'] != 'custom':

                feature_dict['options']['norm_in']['stats'] = stats

            norm_in = Normalization(combined.shape[1], **feature_dict['options']['norm_in'])

            combined = norm_in(torch.FloatTensor(combined))
            combined = combined.numpy()

        # Add timelag features if specified
        if 'timelag' in feature_dict.get('options', {}):

            lag = feature_dict['options']['timelag']
            combined, combined_lag = self.timelag(combined, lag)

            dataset = {"data": combined, "target": combined_lag}

            if self.input_labels_npy_path:
                labels, labels_lag = self.timelag(labels, lag)
                dataset["labels"] = labels_lag

            if self.input_weights_npy_path:
                weights, weights_lag = self.timelag(weights, lag)
                dataset["weights"] = weights_lag

            return dataset, stats

        else:
            dataset = {"data": combined}

            if self.input_labels_npy_path:
                dataset["labels"] = labels
            if self.input_weights_npy_path:
                dataset["weights"] = weights

            return dataset, stats


# Usage example
# traj = md.load("/home/pzanders/Documents/Simulations/GodMD/domini/1NE4_6NO7_b.dcd",
#                 top="/home/pzanders/Documents/Simulations/GodMD/domini/1NE4_6NO7_b.godmd.pdb")

# featurizer = MDFeaturizer(traj)

# atom_pairs = [(57, 78), (57, 79)]
# atom_triplets = [(57, 78, 79), (57, 78, 80)]
# atom_quadruplets = [(57, 78, 79, 80), (57, 78, 79, 81)]

# feature_dict = {
#     'distances': {"pairs": atom_pairs, "cutoff": 0.5, "periodic": True},
#     'angles':    {"triplets": atom_triplets, "periodic": True},
#     'dihedrals': {"quadruplets": atom_quadruplets, "periodic": True}
# }
# combined, details = featurizer.compute_features(feature_dict)



import os


def get_size(path):
    size = os.path.getsize(path)
    if size < 1024:
        return f"{size} bytes"
    elif size < pow(1024, 2):
        return f"{round(size / 1024, 2)} KB"
    elif size < pow(1024, 3):
        return f"{round(size / (pow(1024, 2)), 2)} MB"
    elif size < pow(1024, 4):
        return f"{round(size / (pow(1024, 3)), 2)} GB"



from sklearn.feature_selection import mutual_info_regression
import pandas as pd


def make_mi_scores(X, y, discrete_features):
    mi_scores = mutual_info_regression(X, y, discrete_features=discrete_features)
    mi_scores = pd.Series(mi_scores, name="MI Scores", index=X.columns)
    mi_scores = mi_scores.sort_values(ascending=False)
    return mi_scores

# mi_scores = make_mi_scores(X_pca, y, discrete_features=False)
# mi_scores



#!/usr/bin/env python

# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
(Weighted) Mean Squared Error (MSE) loss function.
"""

__all__ = ["MSELoss", "mse_loss"]


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

from typing import Optional

import torch


# =============================================================================
# LOSS FUNCTIONS
# =============================================================================


class MSELoss(torch.nn.Module):
    """(Weighted) Mean Square Error"""

    def forward(
        self,
        input: torch.Tensor,
        target: torch.Tensor,
        weights: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Compute the value of the loss function."""
        return mse_loss(input, target, weights)


def mse_loss(
    input: torch.Tensor, target: torch.Tensor, weights: Optional[torch.Tensor] = None
) -> torch.Tensor:
    """(Weighted) Mean Square Error

    Parameters
    ----------
    input : torch.Tensor
        prediction
    target : torch.Tensor
        reference
    weights : torch.Tensor, optional
        sample weights, by default None

    Returns
    -------
    loss: torch.Tensor
        loss function
    """
    # reshape in the correct format (batch, size)
    if input.ndim == 1:
        input = input.unsqueeze(1)
    if target.ndim == 1:
        target = target.unsqueeze(1)
    # take the different
    diff = input - target
    # weight them
    if weights is not None:
        if weights.ndim == 1:
            weights = weights.unsqueeze(1)
        loss = (diff * weights).square().mean()
    else:
        loss = diff.square().mean()
    return loss



import mdtraj as md
import numpy as np
from collections import defaultdict
import itertools


class FeaturesGenerator:
    """
    Class to generate a PLUMED features.dat file from an MDTraj trajectory (single frame reference structure)
    and a selection dictionary specifying the types of features to include (cartesian, distances, angles, dihedrals).

    The reference structure is used to compute positions for pair selection in distances (if cutoff is provided).

    Usage:
    generator = FeaturesGenerator(ref_structure='reference.pdb', selection_dict=your_dict)
    generator.generate(output_file='features.dat')
    """

    def __init__(self, ref_structure: str, selection_dict: dict):
        """
        Initialize with path to reference structure (e.g., PDB file) and selection dictionary.

        Args:
            ref_structure: Path to the reference PDB or other MDTraj-loadable file (provides topology and coordinates).
            selection_dict: Dictionary specifying features, e.g.,
                {
                    'cartesian': {'selection': 'backbone'},
                    'distances': {'selection': 'name CA', 'cutoff': 0.4, 'periodic': True, 'bonded': False},
                    'angles': {'selection': 'backbone', 'periodic': True, 'bonded': True},
                    'dihedrals': {'selection': 'backbone', 'periodic': True, 'bonded': True}
                }
        """
        self.traj = md.load(ref_structure)
        self.top = self.traj.top
        self.selection_dict = selection_dict

    def generate(self, output_file: str = 'features.dat'):
        """
        Generate the features.dat file with PLUMED commands for the specified features.

        Args:
            output_file: Path to output features.dat file.
        """
        lines = []

        # Cartesian positions
        if 'cartesian' in self.selection_dict:
            sel = self.selection_dict['cartesian']['selection']
            atoms = self.top.select(sel)
            for i, atom_idx in enumerate(sorted(atoms)):
                label = f"p{i + 1}"
                lines.append(f"{label}: POSITION ATOM={atom_idx + 1}")

        # Distances
        if 'distances' in self.selection_dict:
            sel_dict = self.selection_dict['distances']
            sel = sel_dict['selection']
            cutoff = sel_dict.get('cutoff', None)  # If no cutoff, perhaps all pairs, but assume required
            if cutoff is None:
                raise ValueError("Cutoff must be provided for distances.")
            periodic = sel_dict.get('periodic', True)
            include_bonded = sel_dict.get('bonded', True)  # But in example False

            atoms = self.top.select(sel)
            if len(atoms) == 0:
                raise ValueError(f"No atoms selected for distances with '{sel}'")

            # Compute pairwise distances to find pairs within cutoff
            pairs_list = list(itertools.combinations(atoms, 2))
            if len(pairs_list) > 0:
                dists = md.compute_distances(self.traj[0:1], pairs_list, periodic=periodic)[0]
                pairs = set()
                for ii in range(len(dists)):
                    if dists[ii] < cutoff:
                        a1, a2 = pairs_list[ii]
                        pairs.add((min(a1, a2), max(a1, a2)))
            else:
                pairs = set()

            # Exclude bonded if not include_bonded
            if not include_bonded:
                bond_set = {frozenset({b.atom1.index, b.atom2.index}) for b in self.top.bonds}
                pairs = {p for p in pairs if frozenset(p) not in bond_set}

            pbc_str = " NOPBC" if not periodic else ""
            for i, (a1, a2) in enumerate(sorted(pairs)):
                label = f"d{i + 1}"
                lines.append(f"{label}: DISTANCE ATOMS={a1 + 1},{a2 + 1}{pbc_str}")

        # Angles
        if 'angles' in self.selection_dict:
            sel_dict = self.selection_dict['angles']
            sel = sel_dict['selection']
            bonded = sel_dict.get('bonded', True)
            if not bonded:
                raise NotImplementedError("Non-bonded angles not supported; too many combinations.")

            sel_atoms = set(self.top.select(sel))

            # Build adjacency list for selected atoms
            adj = defaultdict(list)
            for bond in self.top.bonds:
                a1, a2 = bond.atom1.index, bond.atom2.index
                if a1 in sel_atoms and a2 in sel_atoms:
                    adj[a1].append(a2)
                    adj[a2].append(a1)

            # Find triplets
            triplets = set()
            for j in sorted(sel_atoms):
                neigh = sorted(adj[j])
                for idx1 in range(len(neigh)):
                    for idx2 in range(idx1 + 1, len(neigh)):
                        i, k = neigh[idx1], neigh[idx2]
                        ordered = sorted([i, j, k])
                        triplets.add(tuple(ordered))

            for i, triplet in enumerate(sorted(triplets)):
                a1, a2, a3 = triplet
                label = f"a{i + 1}"
                lines.append(f"{label}: ANGLE ATOMS={a1 + 1},{a2 + 1},{a3 + 1}")

        # Dihedrals
        if 'dihedrals' in self.selection_dict:
            sel_dict = self.selection_dict['dihedrals']
            sel = sel_dict['selection']
            bonded = sel_dict.get('bonded', True)
            if not bonded:
                raise NotImplementedError("Non-bonded dihedrals not supported; too many combinations.")

            # Use MDTraj's built-in for backbone dihedrals (phi, psi, omega)
            # Assumes selection is 'backbone' for protein; extend if needed
            phi_indices = md.compute_phi(self.traj)[0]
            psi_indices = md.compute_psi(self.traj)[0]
            omega_indices = md.compute_omega(self.traj)[0]

            all_indices = np.vstack((phi_indices, psi_indices, omega_indices))
            unique_dihedrals = set(tuple(row) for row in all_indices if all(idx in self.top.select(sel) for idx in row))

            for i, dihedral in enumerate(sorted(unique_dihedrals)):
                a1, a2, a3, a4 = dihedral
                label = f"t{i + 1}"
                lines.append(f"{label}: TORSION ATOMS={a1 + 1},{a2 + 1},{a3 + 1},{a4 + 1}")

        # Write to file
        with open(output_file, 'w') as f:
            for line in lines:
                f.write(line + '\n')

        print(f"Features file generated at {output_file}")



# Copyright (c) 2021 Venkata K. Ramaswamy, Samuel C. Musson, Chris G. Willcocks, Matteo T. Degiacomi
#
# Molearn is free software ;
# you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation ;
# either version 2 of the License, or (at your option) any later version.
# molearn is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with molearn ;
# if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.

import torch
# import torch.nn as nn
# import torch.nn.functional as F
from .torch_protein_energy_utils import get_convolutions


class TorchProteinEnergy():
    def __init__(self, frame, pdb_atom_names,
                 padded_residues=False,
                 method=('indexed', 'convolutional', 'roll')[2],
                 device=torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu'), fix_h=False, alt_vdw=[], NB='repulsive'):
        '''
        At instantiation will load amber parameters and create the necessary convolutions/indexs/rolls to calculate the energy of the molecule. Energy can be assessed with the
        ``TorchProteinEnergy.get_energy(x)`` method

        :param frame: Example coordinates of the structure in a torch array. The interatomic distance will be used to determine the connectivity of the atoms.
            Coordinates should be of ``shape [3, N]`` where N is the number of atoms.
            If ``padded_residues = True`` then Coordinates should be of ``shape [R, M, 3]`` where R is the number of residues
            and M is the maximum number of atoms in a residue.
        :param pdb_atom_names: Array of ``shape [N, 2]`` containing the pdb atom names in ``pdb_atom_names[:, 0]`` and residue names in ``pdb_atom_names[:, 1]``. If
            ``padded_residues = True`` then should be ``shape [R, M, 2]``.
        :param padded_residues: If true the dataset should be formatted as ``shape [R, M, 3]`` where R is the number of residues and M is the maximum number of atoms.
            Padding should be ``nan``.
            **Note** only ``method = "indexed"`` is currently implemented currently for this.
        :param method: ``method = "convolutional"`` (currently experimental) method uses convolutions to calculate force (padded_residues=false only).

            ``method = "roll"`` method uses rolling and slicing to calculate force (padded_residues = false only)

            ``method = "indexed"`` (experimental) method is only impremented for padded_residues=True. Uses indexes to calculate forces.
        :param device: ``torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")``
        '''
        self.device = device
        self.method = method
        if padded_residues:
            if method == 'indexed':
                self._padded_indexed_init(frame, pdb_atom_names)
        else:
            if method == 'convolutional':
                self._convolutional_init(frame, pdb_atom_names, fix_h=fix_h, alt_vdw=alt_vdw)
            elif method == 'roll':
                self._roll_init(frame, pdb_atom_names, NB=NB, fix_h=fix_h, alt_vdw=alt_vdw)

    def get_energy(self, x, nonbonded=False):
        '''
            :param x: tensor of shape [B, 3, N] where B is batch_size, N is number of atoms. If padded_residues = True then tensor of shape [B, R, M, 3] where B is R is number of Residues and M is
                the maximum number of atoms in a residue.
            :param nonbonded: Boolean (default False), Whether to return a softened nonbonded energy
            :returns: ``float`` Bond energy (average over batch)
            :returns: ``float`` Angle energy (average over batch)
            :returns: ``float`` Torsion energy (average over batch)
            :returns: ``float`` (optional if nonbonded = True) non-bonded energy (average over batch)
        '''
        if nonbonded:
            return (*self.get_bonded_energy(x), self._nb_loss(x))
        else:
            return self.get_bonded_energy(x)

    def _roll_init(self, frame, pdb_atom_names, NB='full', fix_h=False, alt_vdw=[]):
        (b_masks, b_equil, b_force, b_weights,
            a_masks, a_equil, a_force, a_weights,
            t_masks, t_para, t_weights,
            vdw_R, vdw_e, vdw_14R, vdw_14e,
            q1q2, q1q2_14) = get_convolutions(frame, pdb_atom_names, fix_slice_method=True, fix_h=fix_h, alt_vdw=alt_vdw)

        self.brdiff = []
        self.br_equil = []
        self.br_force = []
        for i, j in enumerate(b_weights):
            atom1 = j.index(1)
            atom2 = j.index(-1)
            d = j.index(-1) - j.index(1)
            padding = len(j) - 2
            self.brdiff.append(d)
            # b_equil[:,0] is just padding so can roll(-1,1) to get correct padding
            self.br_equil.append(torch.tensor(b_equil[i, padding - 1:]).roll(-1).to(self.device).float())
            self.br_force.append(torch.tensor(b_force[i, padding - 1:]).roll(-1).to(self.device).float())
        self.ardiff = []
        self.arsign = []
        self.arroll = []
        self.ar_equil = []
        self.ar_force = []
        self.ar_masks = []
        for i, j in enumerate(a_weights):
            atom1 = j[0].index(1)
            atom2 = j[0].index(-1)
            atom3 = j[1].index(1)
            diff1 = atom2 - atom1
            diff2 = atom2 - atom3
            padding = len(j[0]) - 3
            self.arroll.append([min(atom1, atom2), min(atom2, atom3)])
            self.ardiff.append([abs(diff1) - 1, abs(diff2) - 1])
            self.arsign.append([diff1 / abs(diff1), diff2 / abs(diff2)])
            self.ar_equil.append(torch.tensor(a_equil[i, padding - 2:]).roll(-2).to(self.device).float())
            self.ar_force.append(torch.tensor(a_force[i, padding - 2:]).roll(-2).to(self.device).float())
        self.trdiff = []
        self.trsign = []
        self.trroll = []
        self.tr_para = []
        for i, j in enumerate(t_weights):
            atom1 = j[0].index(1)     # i-j 0
            atom2 = j[0].index(-1)    # i-j 2
            atom3 = j[1].index(-1)    # j-k 3
            atom4 = j[2].index(1)     # l-k 4
            diff1 = atom2 - atom1       # ij 2
            diff2 = atom3 - atom2       # jk 1
            diff3 = (atom4 - atom3) * -1  # lk 1
            padding = len(j[0]) - 4
            self.trroll.append([min(atom1, atom2), min(atom2, atom3), min(atom3, atom4)])
            self.trsign.append([diff1 / abs(diff1), diff2 / abs(diff2), diff3 / abs(diff3)])
            self.trdiff.append([abs(diff1) - 1, abs(diff2) - 1, abs(diff3) - 1])
            self.tr_para.append(torch.tensor(t_para[i, padding - 3:]).roll(-3, 0).to(self.device).float())

        self.vdw_A = (vdw_e * (vdw_R**12)).to(self.device)
        self.vdw_B = (2 * vdw_e * (vdw_R**6)).to(self.device)
        self.q1q2 = q1q2.to(self.device)

        self.get_bonded_energy = self._bonded_roll_loss
        if NB == 'full':
            self._nb_loss = self._cdist_nb_full
        elif NB == 'repulsive':
            self._nb_loss = self._cdist_nb

    def _convolutional_init(self, frame, pdb_atom_names, NB='full', fix_h=False, alt_vdw=[]):
        (b_masks, b_equil, b_force, b_weights,
         a_masks, a_equil, a_force, a_weights,
         t_masks, t_para, t_weights,
         vdw_R, vdw_e, vdw_14R, vdw_14e,
         q1q2, q1q2_14) = get_convolutions(frame, pdb_atom_names, fix_slice_method=False, fix_h=fix_h, alt_vdw=alt_vdw)

        self.b_equil = torch.tensor(b_equil).to(self.device)
        self.b_force = torch.tensor(b_force).to(self.device)
        self.b_weights = torch.tensor(b_weights).to(self.device)

        self.a_equil = torch.tensor(a_equil).to(self.device).float()
        self.a_force = torch.tensor(a_force).to(self.device).float()
        self.a_weights = torch.tensor(a_weights).to(self.device)
        self.a_masks = torch.tensor(a_masks).to(self.device)

        self.t_para = torch.tensor(t_para).to(self.device)
        self.t_weights = torch.tensor(t_weights).to(self.device)

        self.vdw_A = (vdw_e * (vdw_R**12)).to(self.device)
        self.vdw_B = (2 * vdw_e * (vdw_R**6)).to(self.device)
        self.q1q2 = q1q2.to(self.device)

        self.get_bonded_energy = self._bonded_convolutional_loss
        if NB == 'full':
            self._nb_loss = self._cdist_nb_full
        elif NB == 'repulsive':
            self._nb_loss = self._cdist_nb

    def _padded_indexed_init(self, frame, pdb_atom_names, NB='full'):
        from molearn import get_conv_pad_res
        (bond_idxs, bond_para,
         angle_idxs, angle_para, angle_mask, ij_jk,
         torsion_idxs, torsion_para, torsion_mask, ij_jk_kl,
         vdw_R, vdw_e,
         q1q2,) = get_conv_pad_res(frame, pdb_atom_names)

        self.bond_idxs = bond_idxs.to(self.device)
        self.bond_para = bond_para.to(self.device)
        self.angle_mask = angle_mask.to(self.device)
        self.ij_jk = ij_jk.to(self.device)
        self.angle_para = angle_para.to(self.device)
        self.torsion_mask = torsion_mask.to(self.device)
        self.ij_jk_kl = ij_jk_kl.to(self.device)
        self.torsion_para = torsion_para.to(self.device)
        self.vdw_A = (vdw_e * (vdw_R**12)).to(self.device)
        self.vdw_B = (2 * vdw_e * (vdw_R**6)).to(self.device)
        self.q1q2 = q1q2.to(self.device)
        self.get_bonded_energy = self._bonded_padded_residues_loss
        self.relevant = self.bond_idxs.unique().to(self.device)
        if NB == 'full':
            self._nb_loss = self._cdist_nb_full
        elif NB == 'repulsive':
            self._nb_loss = self._cdist_nb

    def _bonded_convolutional_loss(self, x):
        bs = x.shape[0]
        bloss = self._conv_bond_loss(x)
        aloss = self._conv_angle_loss(x)
        tloss = self._conv_torsion_loss(x)
        return bloss / bs, aloss / bs, tloss / bs

    def _bonded_roll_loss(self, x):
        bs = x.shape[0]
        bloss, aloss, tloss = self._roll_bond_angle_torsion_loss(x)
        return bloss / bs, aloss / bs, tloss / bs

    def _bonded_padded_residues_loss(self, x):
        # x.shape [B, R, M, 3]
        x = x.view(x.shape[0], -1, 3)[:,]
        v = x[:, self.bond_idxs[:, 1]] - x[:, self.bond_idxs[:, 0]]  # j-i == i->j
        bloss = (((v.norm(dim=2) - self.bond_para[:, 0])**2) * self.bond_para[:, 1]).sum()
        v1 = v[:, self.ij_jk[0]] * self.angle_mask[0].view(1, -1, 1)
        v2 = v[:, self.ij_jk[1]] * self.angle_mask[1].view(1, -1, 1)
        xyz = torch.sum(v1 * v2, dim=2) / (torch.norm(v1, dim=2) * torch.norm(v2, dim=2))
        theta = torch.acos(torch.clamp(xyz, min=-0.999999, max=0.999999))
        aloss = (((theta - self.angle_para[:, 0])**2) * self.angle_para[:, 1]).sum()

        u1 = v[:, self.ij_jk_kl[0]] * self.torsion_mask[0].view(1, -1, 1)
        u2 = v[:, self.ij_jk_kl[1]] * self.torsion_mask[1].view(1, -1, 1)
        u3 = v[:, self.ij_jk_kl[2]] * self.torsion_mask[2].view(1, -1, 1)
        u12 = torch.cross(u1, u2)
        u23 = torch.cross(u2, u3)
        t3 = torch.atan2(u2.norm(dim=2) * ((u1 * u23).sum(dim=2)), (u12 * u23).sum(dim=2))
        p = self.torsion_para
        tloss = ((p[:, 1] / p[:, 0]) * (1 + torch.cos((p[:, 3] * t3.unsqueeze(2)) - p[:, 2]))).sum()
        bs = x.shape[0]
        return bloss / bs, aloss / bs, tloss / bs

    def _cdist_nb_full(self, x, cutoff=9.0, mask=False):
        dmat = torch.cdist(x.permute(0, 2, 1), x.permute(0, 2, 1))
        dmat6 = (self._warp_domain(dmat, 1.9)**6)
        LJpB = self.vdw_B / dmat6
        LJpA = self.vdw_A / (dmat6**2)
        Cp = (self.q1q2 / self._warp_domain(dmat, 0.4))
        return torch.nansum(LJpA - LJpB + Cp)

    def _cdist_nb(self, x, cutoff=9.0, mask=False):
        dmat = torch.cdist(x.permute(0, 2, 1), x.permute(0, 2, 1))
        LJp = self.vdw_A / (self._warp_domain(dmat, 1.9)**12)
        Cp = (self.q1q2 / self._warp_domain(dmat, 0.4))
        return torch.nansum(LJp + Cp)

    def _warp_domain(self, x, k):
        return torch.nn.functional.elu(x - k, 1.0) + k

    def _conv_bond_loss(self, x):
        # x shape[B, 3, N]
        loss = torch.tensor(0.0).float().to(self.device)
        for i, weight in enumerate(self.b_weights):
            y = torch.nn.functional.conv1d(x, weight.view(1, 1, -1).repeat(3, 1, 1).to(self.device), groups=3, padding=(len(weight) - 2))
            loss += (self.b_force[i] * ((y.norm(dim=1) - self.b_equil[i])**2)).sum()
        return loss

    def _conv_angle_loss(self, x):
        # x shape[X, 3, N]
        loss = torch.tensor(0.0).float().to(self.device)
        for i, weight in enumerate(self.a_weights):
            v1 = torch.nn.functional.conv1d(x, weight[0].view(1, 1, -1).repeat(3, 1, 1).to(self.device), groups=3, padding=(len(weight[0]) - 3))
            v2 = torch.nn.functional.conv1d(x, weight[1].view(1, 1, -1).repeat(3, 1, 1).to(self.device), groups=3, padding=(len(weight[1]) - 3))
            xyz = torch.sum(v1 * v2, dim=1) / (torch.norm(v1, dim=1) * torch.norm(v2, dim=1))
            theta = torch.acos(torch.clamp(xyz, min=-0.999999, max=0.999999))
            energy = (self.a_force[i] * ((theta - self.a_equil[i])**2)).sum(dim=0)[self.a_masks[i]].sum()
            loss += energy
        return loss

    def _conv_torsion_loss(self, x):
        # x shape[X, 3, N]
        loss = torch.tensor(0.0).float().to(self.device)
        for i, weight in enumerate(self.t_weights):
            b1 = torch.nn.functional.conv1d(x, weight[0].view(1, 1, -1).repeat(3, 1, 1).to(self.device), groups=3, padding=(len(weight[0]) - 4))  # i-j
            b2 = torch.nn.functional.conv1d(x, weight[1].view(1, 1, -1).repeat(3, 1, 1).to(self.device), groups=3, padding=(len(weight[1]) - 4))  # j-k
            b3 = torch.nn.functional.conv1d(x, weight[2].view(1, 1, -1).repeat(3, 1, 1).to(self.device), groups=3, padding=(len(weight[2]) - 4))  # l-k
            c32 = torch.cross(b3, b2)
            c12 = torch.cross(b1, b2)
            torsion = torch.atan2((b2 * torch.cross(c32, c12)).sum(dim=1),
                                  b2.norm(dim=1) * ((c12 * c32).sum(dim=1)))
            p = self.t_para[i, :, :, :].unsqueeze(0)
            loss += ((p[:, :, 1] / p[:, :, 0]) * (1 + torch.cos((p[:, :, 3] * torsion.unsqueeze(2)) - p[:, :, 2]))).sum()
        return loss

    def _roll_bond_angle_torsion_loss(self, x):
        # x.shape [5,3,2145]
        bloss = torch.tensor(0.0).float().to(self.device)
        aloss = torch.tensor(0.0).float().to(self.device)
        tloss = torch.tensor(0.0).float().to(self.device)
        v = []
        for i, diff in enumerate(self.brdiff):
            v.append(x - x.roll(-diff, 2))
            bloss += (((v[-1].norm(dim=1) - self.br_equil[i])**2) * self.br_force[i]).sum()

        for i, diff in enumerate(self.ardiff):
            v1 = self.arsign[i][0] * (v[diff[0]].roll(-self.arroll[i][0], 2))
            v2 = self.arsign[i][1] * (v[diff[1]].roll(-self.arroll[i][1], 2))
            xyz = torch.sum(v1 * v2, dim=1) / (torch.norm(v1, dim=1) * torch.norm(v2, dim=1))
            theta = torch.acos(torch.clamp(xyz, min=-0.999999, max=0.999999))
            energy = (self.ar_force[i] * ((theta - self.ar_equil[i])**2))
            sum_e = energy.sum()
            aloss += (sum_e)

        for i, diff in enumerate(self.trdiff):
            b1 = self.trsign[i][0] * (v[diff[0]].roll(-self.trroll[i][0], 2))
            b2 = self.trsign[i][1] * (v[diff[1]].roll(-self.trroll[i][1], 2))
            b3 = self.trsign[i][2] * (v[diff[2]].roll(-self.trroll[i][2], 2))
            c32 = torch.cross(b3, b2)
            c12 = torch.cross(b1, b2)
            torsion = torch.atan2((b2 * torch.cross(c32, c12)).sum(dim=1),
                                  b2.norm(dim=1) * ((c12 * c32).sum(dim=1)))
            p = self.tr_para[i].unsqueeze(0)
            tloss += (((p[:, :, 1] / p[:, :, 0]) * (1 + torch.cos((p[:, :, 3] * torsion.unsqueeze(2)) - p[:, :, 2]))).sum())
        return bloss, aloss, tloss



import os
from typing import Union, List
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.markers import MarkerStyle


def plot_loss(output_train_data_npz_path: str) -> None:
    """
    Plot the training and validation losses from the given npz file.

    Args:
        output_train_data_npz_path (str): The path to the npz file containing the training and validation losses.
    """
    npz_file = np.load(output_train_data_npz_path)
    train_loss = npz_file["train_loss"]
    val_loss = npz_file.get("valid_loss", None)
    min_train_loss_idx = np.argmin(train_loss)
    min_val_loss_idx = np.argmin(val_loss) if val_loss is not None else None

    plt.plot(
        range(len(train_loss)),
        train_loss,
        label=f"Training (min.: {min_train_loss_idx})",
        color="blue",
    )
    if val_loss is not None:
        plt.plot(
            range(len(val_loss)),
            val_loss,
            label=f"Validation (min.: {min_val_loss_idx})",
            color="orange",
        )
    plt.scatter(
        min_train_loss_idx,
        train_loss[min_train_loss_idx],
        color="blue",
        marker=MarkerStyle("v"),
        s=50,
    )
    if val_loss is not None and min_val_loss_idx is not None:
        plt.scatter(
            min_val_loss_idx,
            val_loss[min_val_loss_idx],
            color="orange",
            marker=MarkerStyle("v"),
            s=50,
        )
    plt.legend()
    plt.ylabel("Total Loss")
    plt.xlabel("Epochs")
    plt.title("Training/Validation")
    plt.show()


def plot_rmsd(input_xvg_path: Union[str, List[str]]) -> None:
    """
    Plots RMSD from one or more XVG files.

    Parameters:
    input_xvg_path (str or list of str): Path to a single XVG file or list of paths to multiple XVG files.

    The function parses each XVG file, extracts residue numbers and RMSD values,
    and plots them on a single figure for comparison.
    """
    if isinstance(input_xvg_path, str):
        input_xvg_path = [input_xvg_path]

    plt.figure(figsize=(15, 6))

    for file_path in input_xvg_path:
        if not os.path.exists(file_path):
            print(f"Warning: File {file_path} does not exist. Skipping.")
            continue

        # Load data from XVG, skipping comment lines
        data = np.loadtxt(file_path, comments=['#', '@'])

        # Assume column 0: time, column 1: RMSD
        time = data[:, 0]
        rmsd = data[:, 1]

        # Get label from filename
        label = os.path.basename(file_path).replace('.xvg', '')

        plt.plot(time, rmsd, label=label)

    plt.xlabel('time (ns)')
    plt.ylabel('RMSD (nm)')
    plt.legend()
    plt.grid(True)
    plt.show()


def plot_rmsf(input_xvg_path: Union[str, List[str]]) -> None:
    """
    Plots RMSF from one or more XVG files.

    Parameters:
    input_xvg_path (str or list of str): Path to a single XVG file or list of paths to multiple XVG files.

    The function parses each XVG file, extracts residue numbers and RMSF values,
    and plots them on a single figure for comparison.
    """
    if isinstance(input_xvg_path, str):
        input_xvg_path = [input_xvg_path]

    plt.figure(figsize=(15, 6))

    for file_path in input_xvg_path:
        if not os.path.exists(file_path):
            print(f"Warning: File {file_path} does not exist. Skipping.")
            continue

        # Load data from XVG, skipping comment lines
        data = np.loadtxt(file_path, comments=['#', '@'])

        # Assume column 0: residue, column 1: RMSF
        residues = data[:, 0]
        rmsf = data[:, 1]

        # Get label from filename
        label = os.path.basename(file_path).replace('.xvg', '')

        plt.plot(residues, rmsf, label=label)

    plt.xlabel('Residue Number')
    plt.ylabel('RMSF (nm)')
    plt.legend()
    plt.grid(True)
    plt.show()


def plot_rmsf_difference(input_xvg_path: Union[str, List[str]]) -> None:
    """
    Plots RMSF from one or more XVG files.

    Parameters:
    input_xvg_path (str or list of str): Path to a single XVG file or list of paths to multiple XVG files.

    The function parses each XVG file, extracts residue numbers and RMSF values,
    and plots them on a single figure for comparison.
    """
    if isinstance(input_xvg_path, str):
        input_xvg_path = [input_xvg_path]

    rmsfs = []
    for file_path in input_xvg_path:
        if not os.path.exists(file_path):
            print(f"Warning: File {file_path} does not exist. Skipping.")
            continue

        # Load data from XVG, skipping comment lines
        data = np.loadtxt(file_path, comments=['#', '@'])

        # Assume column 0: residue, column 1: RMSF
        residues = data[:, 0]
        rmsf = data[:, 1]

        # Get label from filename
        label = f"DIO: {os.path.basename(file_path).replace('xvg', '')} vs {os.path.basename(file_path).replace('xvg', '')}"

        rmsfs.append(rmsf)

    diff_rmsf = abs(rmsfs[0] - rmsfs[1])

    plt.figure(figsize=(10, 6))
    plt.plot(residues, diff_rmsf, label=label)
    plt.xlabel('Residue Number')
    plt.ylabel('RMSF (nm)')
    plt.title('RMSF Difference')
    plt.legend()
    plt.grid(True)
    plt.show()


def plot_latent_space(results_npz_path: str,
                      projection_dim: list,
                      snapshot_freq_ps=10) -> None:

    results = np.load(results_npz_path, allow_pickle=True)

    if 'z' not in results:
        raise KeyError(f"'z' not found in {results_npz_path}")

    z = results['z']

    if projection_dim is None:
        projection_dim = [0, 1]

    if len(projection_dim) != 2:
        raise ValueError(f"projection_dim must have length 2, got {projection_dim}")

    dim1, dim2 = projection_dim
    n_frames = z.shape[0]
    n_ticks = int(n_frames / 10)
    timestep_ns = 1 / snapshot_freq_ps

    plt.figure(figsize=(10, 6))
    plt.scatter(z[:, dim1], z[:, dim2], c=np.arange(n_frames) * timestep_ns, s=10, alpha=1.0)
    plt.xlabel(f'latent_dim {dim1}')
    plt.ylabel(f'latent_dim {dim2}')

    ticks = np.arange(n_frames)[::n_ticks] * timestep_ns
    if 0 not in ticks:
        ticks = np.insert(ticks, 0, 0)
    if (n_frames - 1) * timestep_ns not in ticks:
        ticks = np.append(ticks, (n_frames - 1) * timestep_ns)

    plt.colorbar(ticks=ticks, label='Time (ns)')
    plt.title('Latent Space Visualization')
    plt.show()



#!/usr/bin/env python

# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Evidence Lower BOund (ELBO) loss functions used to train variational Autoencoders.
"""

__all__ = ["ELBOGaussiansLoss", "elbo_gaussians_loss", "ELBOLoss", "ELBOGaussianMixtureLoss"]


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

from typing import Optional
import torch
import math
from torch import nn
from torch.nn import functional as F
from mlcolvar.core.loss.mse import mse_loss


# =============================================================================
# LOSS FUNCTIONS
# =============================================================================


class ELBOGaussiansLoss(torch.nn.Module):
    """ELBO loss function assuming the latent and reconstruction distributions are Gaussian.

    The ELBO uses the MSE as the reconstruction loss (i.e., assumes that the
    decoder outputs the mean of a Gaussian distribution with variance 1), and
    the KL divergence between two normal distributions ``N(mean, var)`` and
    ``N(0, 1)``, where ``mean`` and ``var`` are the output of the encoder.
    """

    def forward(
        self,
        target: torch.Tensor,
        output: torch.Tensor,
        mean: torch.Tensor,
        log_variance: torch.Tensor,
        weights: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Compute the value of the loss function.

        Parameters
        ----------
        target : torch.Tensor
            Shape ``(n_batches, in_features)``. Data points (e.g. input of encoder
            or time-lagged features).
        output : torch.Tensor
            Shape ``(n_batches, in_features)``. Output of the decoder.
        mean : torch.Tensor
            Shape ``(n_batches, latent_features)``. The means of the Gaussian
            distributions associated to the inputs.
        log_variance : torch.Tensor
            Shape ``(n_batches, latent_features)``. The logarithm of the variances
            of the Gaussian distributions associated to the inputs.
        weights : torch.Tensor, optional
            Shape ``(n_batches,)`` or ``(n_batches,1)``. If given, the average over
            batches is weighted. The default (``None``) is unweighted.

        Returns
        -------
        loss: torch.Tensor
            The value of the loss function.
        """
        return elbo_gaussians_loss(target, output, mean, log_variance, weights)


def elbo_gaussians_loss(
    target: torch.Tensor,
    output: torch.Tensor,
    mean: torch.Tensor,
    log_variance: torch.Tensor,
    weights: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """ELBO loss function assuming the latent and reconstruction distributions are Gaussian.

    The ELBO uses the MSE as the reconstruction loss (i.e., assumes that the
    decoder outputs the mean of a Gaussian distribution with variance 1), and
    the KL divergence between two normal distributions ``N(mean, var)`` and
    ``N(0, 1)``, where ``mean`` and ``var`` are the output of the encoder.

    Parameters
    ----------
    target : torch.Tensor
        Shape ``(n_batches, in_features)``. Data points (e.g. input of encoder
        or time-lagged features).
    output : torch.Tensor
        Shape ``(n_batches, in_features)``. Output of the decoder.
    mean : torch.Tensor
        Shape ``(n_batches, latent_features)``. The means of the Gaussian
        distributions associated to the inputs.
    log_variance : torch.Tensor
        Shape ``(n_batches, latent_features)``. The logarithm of the variances
        of the Gaussian distributions associated to the inputs.
    weights : torch.Tensor, optional
        Shape ``(n_batches,)`` or ``(n_batches,1)``. If given, the average over
        batches is weighted. The default (``None``) is unweighted.

    Returns
    -------
    loss: torch.Tensor
        The value of the loss function.
    """
    # KL divergence between N(mean, variance) and N(0, 1).
    # See https://stats.stackexchange.com/questions/7440/kl-divergence-between-two-univariate-gaussians
    kl = -0.5 * (log_variance - log_variance.exp() - mean**2 + 1).sum(dim=1)

    # Weighted mean over batches.
    if weights is None:
        kl = kl.mean()
    else:
        weights = weights.squeeze()
        if weights.shape != kl.shape:
            raise ValueError(
                f"weights should be a tensor of shape (n_batches,) or (n_batches,1), not {weights.shape}."
            )
        kl = (kl * weights).sum()

    # Reconstruction loss.
    reconstruction = mse_loss(output, target, weights=weights)

    return reconstruction + kl


class ELBOLoss(nn.Module):
    """
    Variational Autoencoder ELBO loss function.

    Implements the evidence lower bound (ELBO) objective:
        L = reconstruction_loss + beta * KL_divergence

    Reconstruction loss options:
      - Mean-squared error (MSE) -> assumes Gaussian decoder with unit variance
      - Binary cross-entropy (BCE) -> assumes Bernoulli decoder

    KL divergence is computed analytically between the approximate posterior
    q(z|x) = N(mu, diag(var)) and the prior p(z) = N(0, I):
        KL(q||p) = -0.5 * sum(1 + log(var) - mu^2 - var)

    Parameters
    ----------
    beta : float, default=1.0
        Scaling factor for the KL divergence term (beta-VAE).
    loss_type : {'mse', 'bce'}, default='mse'
        Type of reconstruction loss:
        - 'mse': use mean squared error
        - 'bce': use binary cross-entropy
    reduction : {'sum', 'mean', 'none'}, default='sum'
        How to reduce the reconstruction loss over elements:
        - 'sum': sum over all elements
        - 'mean': average over all elements
        - 'none': no reduction (returns per-element loss)
    """

    def __init__(
        self,
        beta: float = 1.0,
        reconstruction: str = 'mse',
        reduction: str = 'sum'
    ):
        super().__init__()
        if reconstruction not in {'mse', 'bce'}:
            raise ValueError(f"Unsupported reconstruction '{reconstruction}', choose 'mse' or 'bce'.")
        if reduction not in {'sum', 'mean', 'none'}:
            raise ValueError(f"Unsupported reduction '{reduction}', choose 'sum', 'mean', or 'none'.")

        self.beta = beta
        self.reconstruction = reconstruction
        self.reduction = reduction

    def forward(
        self,
        x: torch.Tensor,
        recon_x: torch.Tensor,
        mu: torch.Tensor,
        log_var: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute the combined ELBO loss.

        Parameters
        ----------
        x : Tensor
            Original input tensor (shape: [batch_size, ...]).
        recon_x : Tensor
            Reconstructed output tensor (same shape as x).
        mu : Tensor
            Mean of the approximate posterior q(z|x) (shape: [batch_size, latent_dim]).
        log_var : Tensor
            Log-variance of q(z|x) (same shape as mu).

        Returns
        -------
        loss : Tensor
            Scalar loss (if reduction!='none') or tensor of per-element losses.
        """
        # Reconstruction loss
        if self.reconstruction == 'bce':
            # For binary data, use BCE
            recon_loss = F.binary_cross_entropy(
                recon_x, x, reduction=self.reduction
            )
        else:
            # For continuous data, use MSE
            recon_loss = F.mse_loss(
                recon_x, x, reduction=self.reduction
            )

        # Analytic KL divergence between N(mu, var) and N(0, I)
        # var = exp(log_var)
        var = torch.exp(log_var)
        kl_div = -0.5 * torch.sum(
            1 + log_var - mu.pow(2) - var,
            dim=1  # sum over latent dimension for each sample
        )

        # Combine terms: sum or mean over batch
        if self.reduction == 'mean':
            kl_div = kl_div.mean()
        elif self.reduction == 'sum':
            kl_div = kl_div.sum()
        # else 'none': keep per-sample KL vector

        # Scale KL and add reconstruction
        return recon_loss + self.beta * kl_div


class ELBOGaussianMixtureLoss(nn.Module):
    """
    Gaussian Mixture VAE loss.

    Combines:
      1) Entropy regularization:  -∑_i q(y=i|x) log q(y=i|x)
      2) Reconstruction + KL:
         - E_{q(y|x)} [ log p(x|z,y) ]
         + E_{q(y|x)} [ KL( q(z|x,y) ‖ p(z|y) ) ]
    """

    def __init__(self, k: int, r_nent: float = 1.0):
        """
        Args:
            k       Number of mixture components.
            r_nent  Weight on the entropy term.
        """
        super().__init__()
        self.k = k
        self.r_nent = r_nent

    @staticmethod
    def log_normal(x: torch.Tensor,
                   mu: torch.Tensor,
                   var: torch.Tensor,
                   eps: float = 1e-10) -> torch.Tensor:
        """
        Compute log N(x; mu, var) summed over the last dim:
          -½ ∑ [ log(2π) + (x−μ)^2 / var + log var ]
        """
        const = math.log(2 * math.pi)
        return -0.5 * torch.sum(
            const + (x - mu).pow(2) / (var + eps) + var.log(),
            dim=-1
        )

    def forward(self,
                x: torch.Tensor,
                qy_logit: torch.Tensor,
                xm_list: list[torch.Tensor],
                xv_list: list[torch.Tensor],
                z_list: list[torch.Tensor],
                zm_list: list[torch.Tensor],
                zv_list: list[torch.Tensor],
                zm_prior_list: list[torch.Tensor],
                zv_prior_list: list[torch.Tensor]
                ) -> torch.Tensor:
        """
        Args:
            x                [batch, n_features]                Input data
            qy_logit         [batch, k]                          Cluster logits
            xm_list, xv_list length-k lists of [batch, n_features]
            z_list, zm_list, zv_list       length-k lists of [batch, n_cvs]
            zm_prior_list, zv_prior_list   length-k lists of [batch, n_cvs]
        Returns:
            scalar loss = mean_batch( r_nent*nent + ∑_i qy_i * [rec_i + KL_i] )
        """
        # 1) cluster posteriors
        qy = F.softmax(qy_logit, dim=1)    # [batch, k]

        # 2) entropy regularization (cross-entropy of qy wrt itself)
        #    nent = -E[ log q(y|x) ]
        nent = -torch.sum(qy * F.log_softmax(qy_logit, dim=1), dim=1).mean()

        # 3) per-component reconstruction + KL
        comp_losses = []
        for i in range(self.k):
            # reconstruction:  - log p(x | z_i)
            rec_i = -self.log_normal(x, xm_list[i], xv_list[i])
            # KL divergence:   KL( q(z|x,y=i) ‖ p(z|y=i) )
            kl_i = (
                self.log_normal(z_list[i], zm_list[i], zv_list[i]) -
                self.log_normal(z_list[i], zm_prior_list[i], zv_prior_list[i])
            )
            comp_losses.append(rec_i + kl_i)  # shape [batch]

        # 4) weight each comp by qy[:,i] and sum
        weighted = [qy[:, i] * comp_losses[i] for i in range(self.k)]
        total = self.r_nent * nent + sum(weighted)  # shape [batch]

        return total, nent



"""
SPIB: A deep learning-based framework to learn RCs
from MD trajectories. Code maintained by Dedi.

Read and cite the following when using this method:
https://aip.scitation.org/doi/abs/10.1063/5.0038198
"""

# --------------------
# Model
# --------------------

import torch
import torch.nn as nn
import torch.nn.functional as F
import lightning.pytorch as pl
from mlcolvar.cvs import BaseCV
from biobb_pytorch.mdae.models.nn.feedforward import FeedForward
from biobb_pytorch.mdae.loss import InformationBottleneckLoss
from typing import Optional

__all__ = ["SPIB"]


class SPIB(BaseCV, pl.LightningModule):
    BLOCKS = ["norm_in", "encoder", "decoder", "k",
              "UpdateLabel", "beta", "threshold", "patience", "refinements",
              "learning_rate", "lr_step_size", "lr_gamma"]

    def __init__(self, n_features, n_cvs, encoder_layers, decoder_layers, options=None, **kwargs):
        super().__init__(in_features=n_features, out_features=n_cvs, **kwargs)

        options = self.parse_options(options)

        self._n_cvs = n_cvs
        self.k = options.get("k", 2)
        self.output_dim = n_features

        self.learning_rate = options.get("optimizer", {}).get("lr", 0.001)
        self.lr_step_size = options.get("optimizer", {}).get("step_size", 10)
        self.lr_gamma = options.get("optimizer", {}).get("gamma", 0.1)

        self.beta = 0.01
        self.threshold = options.get("threshold", 0.01)
        self.patience = options.get("patience", 10)
        self.refinements = options.get("refinements", 5)

        self.update_times = 0
        self.unchanged_epochs = 0
        self.state_population0 = None
        self.eps = 1e-10

        # Representative inputs
        self.representative_inputs = torch.eye(
            self.k, self.output_dim, device=self.device, requires_grad=False
        )
        self.idle_input = torch.eye(
            self.k, self.k, device=self.device, requires_grad=False
        )
        self.representative_weights = nn.Sequential(
            nn.Linear(self.k, 1, bias=False),
            nn.Softmax(dim=0)
        )
        nn.init.ones_(self.representative_weights[0].weight)

        # Encoder / Decoder
        o = "encoder"
        self.encoder = FeedForward([n_features] + encoder_layers, **options[o])
        self.encoder_mean = torch.nn.Linear(
            in_features=encoder_layers[-1], out_features=n_cvs
        )
        self.encoder_logvar = torch.nn.Linear(
            in_features=encoder_layers[-1], out_features=n_cvs
        )

        o = "decoder"
        self.decoder = FeedForward([n_cvs] + decoder_layers + [n_features], **options[o])

        # IB loss
        self.loss_fn = InformationBottleneckLoss(beta=self.beta, eps=self.eps)

        self.eval_variables = ["xhat", "z", "mu", "logvar", "labels"]

    def encode(self, inputs: torch.Tensor):
        h = self.encoder(inputs)
        mu = self.encoder_mean(h)
        logvar = -10 * F.sigmoid(self.encoder_logvar(h))
        return mu, logvar

    def decode(self, z: torch.Tensor):
        return F.log_softmax(self.decoder(z), dim=1)

    def forward_cv(self, x: torch.Tensor) -> torch.Tensor:
        if self.norm_in is not None:
            x = self.norm_in(x)
        mu, logvar = self.encode(x)
        z = self.reparameterize(mu, logvar)
        return z

    def reparameterize(self, mu: torch.Tensor, logvar: torch.Tensor):
        std = torch.exp(0.5 * logvar)
        return mu + torch.randn_like(std) * std

    def encode_decode(self, x: torch.Tensor):
        flat = x.view(x.size(0), -1)
        if self.norm_in is not None:
            flat = self.norm_in(flat)
        mu, logvar = self.encode(flat)
        z = self.reparameterize(mu, logvar)
        x_hat = self.decode(z)
        if self.norm_in is not None:
            x_hat = self.norm_in.inverse(x_hat)
        return x_hat, z, mu, logvar

    def evaluate_model(self, batch, batch_idx):

        xhat, z, mu, logvar = self.encode_decode(batch['data'])

        pred = xhat.exp()
        labels = pred.argmax(1)

        return xhat, z, mu, logvar, labels

    def configure_optimizers(self):
        opt = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        sch = torch.optim.lr_scheduler.StepLR(opt, step_size=self.lr_step_size, gamma=self.lr_gamma)
        return {'optimizer': opt, 'lr_scheduler': {'scheduler': sch, 'interval': 'epoch'}}

    @torch.no_grad()
    def update_labels(self, inputs: torch.Tensor) -> Optional[torch.Tensor]:
        if not self.UpdateLabel:
            return None
        loader = self.trainer.datamodule.train_dataloader()
        bs = loader.batch_size
        labels = []
        for i in range(0, len(inputs), bs):
            batch = inputs[i:i + bs].to(self.device)
            mu, _ = self.encode(batch)
            logp = self.decode(mu)
            labels.append(logp.exp())
        preds = torch.cat(labels, dim=0)
        idx = preds.argmax(dim=1)
        return F.one_hot(idx, num_classes=self.k)

    @torch.no_grad()
    def get_representative_z(self):
        return self.encode(self.representative_inputs)

    def reset_representative(self, rep_inputs: torch.Tensor):
        self.representative_inputs = rep_inputs.detach().clone()
        dim = rep_inputs.size(0)
        self.idle_input = torch.eye(dim, dim, device=self.device, requires_grad=False)
        self.representative_weights = nn.Sequential(
            nn.Linear(dim, 1, bias=False), nn.Softmax(dim=0)
        )
        nn.init.ones_(self.representative_weights[0].weight)

    def training_step(self, batch, batch_idx):
        x, y = batch['data'], batch['labels']
        w_batch = batch.get('weights', None)
        preds, z, mu, logvar = self.encode_decode(x)
        rep_mu, rep_logvar = self.get_representative_z()
        w = self.representative_weights(self.idle_input)
        loss, recon_err, kl = self.loss_fn(
            y.to(self.device), preds, z, mu, logvar, rep_mu, rep_logvar, w, w_batch
        )
        name = "train" if self.training else "valid"
        self.log(f'{name}_loss', loss, on_step=True, on_epoch=True, prog_bar=True)
        self.log(f'{name}_recon', recon_err, on_step=True, on_epoch=True, prog_bar=True)
        self.log(f'{name}_kl', kl, on_step=True, on_epoch=True, prog_bar=True)
        return loss

    @torch.no_grad()
    def on_train_epoch_start(self):
        if self.trainer.current_epoch == 0:
            ds = self.trainer.datamodule.train_dataloader().dataset
            self.state_population0 = ds['labels'].float().mean(dim=0)
            self.representative_inputs = torch.eye(
                self.k, self.output_dim, device=self.device, requires_grad=False
            )

    @torch.no_grad()
    def on_train_epoch_end(self):
        ds = self.trainer.datamodule.train_dataloader().dataset
        new_labels = self.update_labels(ds['target'])
        if new_labels is None:
            return
        state_pop = new_labels.float().mean(dim=0)
        delta = torch.norm(state_pop - self.state_population0)
        self.log('state_population_change', delta)
        self.state_population0 = state_pop
        if delta < self.threshold:
            self.unchanged_epochs += 1
            if self.unchanged_epochs > self.patience:
                if torch.sum(state_pop > 0) < 2:
                    self.trainer.should_stop = True
                    return
                if self.UpdateLabel and self.update_times < self.refinements:
                    self.update_times += 1
                    self.unchanged_epochs = 0
                    ds['labels'] = new_labels
                    reps = self.estimate_representative_inputs(
                        ds['data'], getattr(ds, 'weights', None)
                    ).to(self.device)
                    self.reset_representative(reps)
                    self.log(f'refinement_{self.update_times}', 1)
                    # Force reload of the DataLoader to reflect updated labels
                    loop = self.trainer.fit_loop
                    loop._combined_loader = None
                    loop.setup_data()
                else:
                    self.trainer.should_stop = True
        else:
            self.unchanged_epochs = 0

    @torch.no_grad()
    def estimate_representative_inputs(
        self, inputs: torch.Tensor, bias: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        self.eval()
        N = len(inputs)
        bs = self.trainer.datamodule.train_dataloader().batch_size
        preds = []
        for i in range(0, N, bs):
            batch = inputs[i:i + bs].to(self.device)
            mu, _ = self.encode(batch)
            logp = self.decoder(mu)
            preds.append(logp.exp())
        preds = torch.cat(preds, dim=0)
        labels = F.one_hot(preds.argmax(dim=1), num_classes=self.k)

        if bias is None:
            bias = torch.ones(N, device=self.device)

        data_shape = inputs.shape[1:]
        state_sums = torch.zeros(self.k, *data_shape, device=self.device)
        state_counts = torch.zeros(self.k, device=self.device)

        for state in range(self.k):
            mask = labels[:, state] == 1
            if mask.any():
                weights_expanded = bias[mask].view(-1, *([1] * len(data_shape)))
                weighted_inputs = inputs[mask] * weights_expanded
                state_sums[state] += weighted_inputs.sum(dim=0)
                state_counts[state] += bias[mask].sum()

        reps = torch.zeros(self.k, *data_shape, device=self.device)
        for state in range(self.k):
            if state_counts[state] > 0:
                reps[state] = state_sums[state] / state_counts[state]

        return reps



from . import build_model
from . import mdfeaturizer
from . import train_model
from . import evaluate_model
from . import decode_model
from . import encode_model
from . import make_plumed
from . import feat2traj
from .explainability import LRP

name = 'mdae'
__all__ = [
    'build_model',
    'mdfeaturizer',
    'train_model',
    'evaluate_model',
    'decode_model',
    'encode_model',
    'LRP',
    'make_plumed',
    'feat2traj',
]



import torch
from torch.utils.data import DataLoader
import os
from biobb_common.tools.file_utils import launchlogger
from biobb_common.tools import file_utils as fu
from biobb_pytorch.mdae.utils.log_utils import get_size
from biobb_common.generic.biobb_object import BiobbObject
from mlcolvar.data import DictDataset
import numpy as np


class EvaluateEncoder(BiobbObject):
    """
    | biobb_pytorch EvaluateEncoder
    | Encode data with a Molecular Dynamics AutoEncoder (MDAE) model.
    | Evaluates a PyTorch autoencoder from the given properties.

    Args:
        input_model_pth_path (str): Path to the trained model file whose encoder will be used. File type: input. `Sample file <https://github.com/bioexcel/biobb_pytorch/raw/master/biobb_pytorch/test/reference/mdae/output_model.pth>`_. Accepted formats: pth (edam:format_2333).
        input_dataset_pt_path (str): Path to the input dataset file (.pt) to encode. File type: input. `Sample file <https://github.com/bioexcel/biobb_pytorch/raw/master/biobb_pytorch/test/reference/mdae/output_model.pt>`_. Accepted formats: pt (edam:format_2333).
        output_results_npz_path (str): Path to the output latent-space results file (compressed NumPy archive, typically containing 'z'). File type: output. `Sample file <https://github.com/bioexcel/biobb_pytorch/raw/master/biobb_pytorch/test/reference/mdae/output_results.npz>`_. Accepted formats: npz (edam:format_2333).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **Dataset** (*dict*) - ({}) mlcolvar DictDataset / DataLoader options (e.g. batch_size, shuffle).

    Examples:
        This example shows how to use the EvaluateEncoder class to evaluate a PyTorch autoencoder model::

            from biobb_pytorch.mdae.evaluate_model import evaluateEncoder

            input_model_pth_path='input_model.pth'
            input_dataset_pt_path='input_dataset.npy'
            output_results_npz_path='output_results.npz'

            prop={
                'Dataset': {
                    'batch_size': 32
                }
            }

            evaluateEncoder(input_model_pth_path=input_model.pth,
                    input_dataset_pt_path=input_dataset.npy,
                    output_results_npz_path=output_results.npz,
                    properties=prop)

    Info:
        * wrapped_software:
            * name: PyTorch
            * version: >=1.6.0
            * license: BSD 3-Clause
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(
        self,
        input_model_pth_path: str,
        input_dataset_pt_path: str,
        output_results_npz_path: str,
        properties: dict,
        **kwargs,
    ) -> None:

        properties = properties or {}

        super().__init__(properties)

        self.input_model_pth_path = input_model_pth_path
        self.input_dataset_pt_path = input_dataset_pt_path
        self.output_results_npz_path = output_results_npz_path
        self.properties = properties.copy()
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {
                "input_model_pth_path": input_model_pth_path,
                "input_dataset_pt_path": input_dataset_pt_path,
            },
            "out": {
                "output_results_npz_path": output_results_npz_path,
            },
        }

        self.Dataset = self.properties.get('Dataset', {})
        self.results = None

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    def load_model(self):
        return torch.load(self.io_dict["in"]["input_model_pth_path"],
                          weights_only=False)

    def load_dataset(self):
        dataset = torch.load(self.io_dict["in"]["input_dataset_pt_path"],
                             weights_only=False)
        return DictDataset(dataset)

    def create_dataloader(self, dataset):
        ds_cfg = self.properties['Dataset']
        return DataLoader(
            dataset,
            batch_size=ds_cfg.get('batch_size', 16),
            shuffle=ds_cfg.get('shuffle', False),
        )

    def evaluate_encoder(self, model, dataloader):
        """Evaluate the encoder part of the model."""
        model.eval()
        with torch.no_grad():
            z_all = []
            for batch in dataloader:
                z = model.forward_cv(batch['data'])
                z_all.append(z)
        return {"z": torch.cat(z_all, dim=0)}

    @launchlogger
    def launch(self) -> int:
        """
        Execute the :class:`EvaluateEncoder` class and its `.launch()` method.
        """

        fu.log('## BioBB Model Evaluator ##', self.out_log)

        # Setup Biobb
        if self.check_restart():
            return 0

        self.stage_files()

        # Start Pipeline

        # load the model
        fu.log(f'Load model from {os.path.abspath(self.io_dict["in"]["input_model_pth_path"])}', self.out_log)
        model = self.load_model()

        # load the dataset
        fu.log(f'Load dataset from {os.path.abspath(self.io_dict["in"]["input_dataset_pt_path"])}', self.out_log)
        dataset = self.load_dataset()

        # create the dataloader
        fu.log('Start evaluating...', self.out_log)
        dataloader = self.create_dataloader(dataset)

        # evaluate the model
        results = self.evaluate_encoder(model, dataloader)

        # Save the results
        np.savez_compressed(self.io_dict["out"]["output_results_npz_path"], **results)
        fu.log(f'Evaluation Results saved to {os.path.abspath(self.io_dict["out"]["output_results_npz_path"])}', self.out_log)
        fu.log(f'File size: {get_size(self.io_dict["out"]["output_results_npz_path"])}', self.out_log)

        # Copy files to host
        self.copy_to_host()

        # Remove temporal files
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return 0


def evaluateEncoder(
    properties: dict,
    input_model_pth_path: str,
    input_dataset_pt_path: str,
    output_results_npz_path: str,
    **kwargs,
) -> int:
    """Create the :class:`EvaluateEncoder <EvaluateEncoder.EvaluateEncoder>` class and
    execute the :meth:`launch() <EvaluateEncoder.evaluateEncoder.launch>` method."""
    return EvaluateEncoder(**dict(locals())).launch()


evaluateEncoder.__doc__ = EvaluateEncoder.__doc__
main = EvaluateEncoder.get_main(evaluateEncoder, "Encode data with a Molecular Dynamics AutoEncoder (MDAE) model.")

if __name__ == "__main__":
    main()



# import torch
# import lightning
# from mlcolvar.cvs import BaseCV
# from torch_geometric.nn import GCNConv, global_mean_pool
# from mlcolvar.core.transform.utils import Inverse
# from biobb_pytorch.mdae.featurization.normalization import Normalization
# from biobb_pytorch.mdae.loss import MSELoss

# __all__ = ["GNNAutoEncoder"]

# class GNNEncoder(torch.nn.Module):
#     def __init__(self, in_channels, hidden_dim, latent_dim):
#         super().__init__()
#         self.conv1 = GCNConv(in_channels, hidden_dim)
#         self.conv2 = GCNConv(hidden_dim, hidden_dim)
#         self.fc_mu = torch.nn.Linear(hidden_dim, latent_dim)

#     def forward(self, x, edge_index, batch):
#         x = torch.relu(self.conv1(x, edge_index))
#         x = torch.relu(self.conv2(x, edge_index))
#         x = global_mean_pool(x, batch)
#         z = self.fc_mu(x)
#         return z

# class GNNDecoder(torch.nn.Module):
#     def __init__(self, latent_dim, hidden_dim, out_features):
#         super().__init__()
#         self.mlp = torch.nn.Sequential(
#             torch.nn.Linear(latent_dim, hidden_dim),
#             torch.nn.ReLU(),
#             torch.nn.Linear(hidden_dim, hidden_dim),
#             torch.nn.ReLU()
#         )
#         self.out_layer = torch.nn.Linear(hidden_dim, out_features)

#     def forward(self, z, batch):
#         z_expanded = z[batch]
#         x_rec = self.out_layer(self.mlp(z_expanded))
#         return x_rec

# class GNNAutoEncoder(BaseCV, lightning.LightningModule):
#     BLOCKS = ["norm_in", "encoder", "decoder"]

#     def __init__(
#         self,
#         n_cvs: int,
#         encoder_layers: list,
#         decoder_layers: list = None,
#         edge_index: torch.Tensor = None,
#         options: dict = None,
#         **kwargs,
#     ):
#         super().__init__(in_features=encoder_layers[0], out_features=n_cvs, **kwargs)

#         self.loss_fn = MSELoss()
#         options = self.parse_options(options)

#         self.edge_index = edge_index
#         hidden_dim = encoder_layers[1] if len(encoder_layers) > 1 else 64

#         if decoder_layers is None:
#             decoder_layers = encoder_layers[::-1]

#         if options.get("norm_in", True):
#             self.norm_in = Normalization(self.in_features, **options.get("norm_in", {}))
#         else:
#             self.norm_in = None

#         self.encoder = GNNEncoder(in_channels=self.in_features, hidden_dim=hidden_dim, latent_dim=n_cvs)
#         self.decoder = GNNDecoder(latent_dim=n_cvs, hidden_dim=hidden_dim, out_features=self.in_features)

#     def forward_cv(self, x):
#         # x: [B, F], F = in_features
#         if self.norm_in is not None:
#             x = self.norm_in(x)
#         return self.encoder(x, self.edge_index, torch.zeros(x.shape[0], dtype=torch.long, device=x.device))

#     def encode_decode(self, x):
#         B, F = x.shape
#         if self.norm_in is not None:
#             x = self.norm_in(x)
#         # Simulate each frame as one node in a graph
#         z = self.encoder(x, self.edge_index, torch.zeros(B, dtype=torch.long, device=x.device))
#         x_rec = self.decoder(z, torch.arange(B, device=x.device))
#         if self.norm_in is not None:
#             x_rec = self.norm_in.inverse(x_rec)
#         return x_rec

#     def training_step(self, batch, batch_idx):
#         x = batch["data"]
#         target = batch.get("target", x)

#         x_hat = self.encode_decode(x)
#         loss = self.loss_fn(x_hat, target)

#         name = "train" if self.training else "valid"
#         self.log(f"{name}_loss", loss, on_step=True, on_epoch=True, prog_bar=True)
#         return loss

#     def get_decoder(self, return_normalization=False):
#         if return_normalization:
#             if self.norm_in is not None:
#                 inv_norm = Inverse(module=self.norm_in)
#                 decoder_model = torch.nn.Sequential(self.decoder, inv_norm)
#             else:
#                 raise ValueError("return_normalization is set to True but self.norm_in is None")
#         else:
#             decoder_model = self.decoder
#         return decoder_model

#     @property
#     def example_input_array(self):
#         return None



import torch
import torch.nn as nn
import lightning.pytorch as pl
from mlcolvar.cvs import BaseCV
from biobb_pytorch.mdae.featurization.normalization import Normalization
from mlcolvar.core.transform.utils import Inverse
from biobb_pytorch.mdae.models.nn.feedforward import FeedForward
from biobb_pytorch.mdae.loss import ELBOGaussianMixtureLoss


__all__ = ["GaussianMixtureVariationalAutoEncoder"]


class GaussianMixtureVariationalAutoEncoder(BaseCV, pl.LightningModule):
    """Gaussian Mixture Variational AutoEncoder Collective Variable.
    This class implements a Gaussian Mixture Variational AutoEncoder (GMVAE) for
    collective variable (CV) learning. The GMVAE is a generative model that combines
    the principles of Gaussian Mixture Models (GMM) and Variational Autoencoders (VAE).
    It learns a latent representation of the input data by modeling it as a mixture
    of Gaussians, where each Gaussian corresponds to a different cluster in the data.
    The model consists of an encoder that maps the input data to a latent space,
    and a decoder that reconstructs the input data from the latent representation.
    The GMVAE is trained using a variational inference approach, where the model
    learns to maximize the evidence lower bound (ELBO) on the data likelihood.
    The ELBO consists of two terms: the reconstruction loss and the KL divergence
    between the learned latent distribution and a prior distribution (usually a
    standard normal distribution). The GMVAE can be used for various tasks such as
    clustering, dimensionality reduction, and generative modeling.
    The model is designed to work with PyTorch and PyTorch Lightning, making it easy
    to integrate into existing workflows and leverage GPU acceleration.
    Parameters
    ----------
    k : int
        The number of clusters in the Gaussian Mixture Model.
    n_cvs : int
        The dimension of the CV or, equivalently, the dimension of the latent
        space of the autoencoder.
    n_features : int
        The dimension of the input data.
    r_nent : float
        The weight for the entropy regularization term.
    qy_dims : list
        The dimensions of the layers in the encoder for the cluster assignment.
    qz_dims : list
        The dimensions of the layers in the encoder for the latent variable.
    pz_dims : list
        The dimensions of the layers in the decoder for the latent variable.
    px_dims : list
        The dimensions of the layers in the decoder for the reconstruction.
    options : dict, optional
        Additional options for the model, such as normalization and dropout rates.
    """

    BLOCKS = ["norm_in", "encoder", "decoder", "k"]

    def __init__(self, n_features, n_cvs, encoder_layers, decoder_layers, options=None, **kwargs):
        super().__init__(in_features=n_features, out_features=n_cvs, **kwargs)

        options = self.parse_options(options)

        if "norm_in" in options and options["norm_in"] is not None:
            self.norm_in = Normalization(self.in_features, **options["norm_in"])

        self.k = options["k"]
        self.r_nent = options.get('loss_function', {}).get("r_nent", 0.5)

        qy_dims = encoder_layers["qy_dims"]
        qz_dims = encoder_layers["qz_dims"]
        pz_dims = decoder_layers["pz_dims"]
        px_dims = decoder_layers["px_dims"]

        self.loss_fn = ELBOGaussianMixtureLoss(r_nent=self.r_nent, k=self.k)

        self.encoder = nn.ModuleDict()
        self.decoder = nn.ModuleDict()

        self.encoder['y_transform'] = nn.Linear(self.k, self.k)

        self.encoder['qy_nn'] = FeedForward([n_features] + qy_dims + [self.k], **options["encoder"]['qy_nn'])

        self.encoder['qz_nn'] = FeedForward([n_features + self.k] + qz_dims, **options["encoder"]['qz_nn'])
        self.encoder['zm_layer'] = nn.Linear(qz_dims[-1], n_cvs)
        self.encoder['zv_layer'] = nn.Linear(qz_dims[-1], n_cvs)

        self.decoder['pz_nn'] = FeedForward([self.k] + pz_dims, **options["decoder"]['pz_nn'])
        self.decoder['zm_prior_layer'] = nn.Linear(pz_dims[-1], n_cvs)
        self.decoder['zv_prior_layer'] = nn.Linear(pz_dims[-1], n_cvs)

        self.decoder['px_nn'] = FeedForward([n_cvs] + px_dims, **options["decoder"]['px_nn'])
        self.decoder['xm_layer'] = nn.Linear(px_dims[-1], n_features)
        self.decoder['xv_layer'] = nn.Linear(px_dims[-1], n_features)

        self.eval_variables = ["xhat", "z", "qy"]

    @staticmethod
    def log_normal(x, mu, var, eps=1e-10):
        return -0.5 * torch.sum(torch.log(torch.tensor(2.0) * torch.pi) + (x - mu).pow(2) / var + var.log(), dim=-1)  # log probability of a normal (Gaussian) distribution

    def loss_function(self, x, xm, xv, z, zm, zv, zm_prior, zv_prior):
        return (
            -self.log_normal(x, xm, xv) +  # Reconstruction Loss
            self.log_normal(z, zm, zv) - self.log_normal(z, zm_prior, zv_prior) -  # Regularization Loss (KL Divergence)
            torch.log(torch.tensor(1 / self.k, device=x.device))  # Entropy Regularization
        )

    def encode_decode(self, x):

        if self.norm_in is not None:
            data = self.norm_in(x)

        qy_logit = self.encoder['qy_nn'](data)

        y_ = torch.zeros([data.shape[0], self.k]).to(data.device)

        zm_list, zv_list, z_list = [], [], []
        xm_list, xv_list, x_list = [], [], []
        zm_prior_list, zv_prior_list = [], []

        for i in range(self.k):
            # One-hot y
            y = y_ + torch.eye(self.k).to(data.device)[i]

            # Qz
            h0 = self.encoder['y_transform'](y)
            xy = torch.cat([data, h0], dim=1)
            qz_logit = self.encoder['qz_nn'](xy)
            zm = self.encoder['zm_layer'](qz_logit)
            zv = torch.nn.functional.softplus(self.encoder['zv_layer'](qz_logit))
            noise = torch.randn_like(torch.sqrt(zv))
            z_sample = zm + noise * zv

            zm_list.append(zm)
            zv_list.append(zv)
            z_list.append(z_sample)

            # Pz (prior)
            pz_logit = self.decoder['pz_nn'](y)
            zm_prior = self.decoder['zm_prior_layer'](pz_logit)
            zv_prior = torch.nn.functional.softplus(self.decoder['zv_prior_layer'](pz_logit))
            noise = torch.randn_like(torch.sqrt(zv_prior))
            z_prior_sample = zm_prior + noise * zv_prior

            zm_prior_list.append(zm_prior)
            zv_prior_list.append(zv_prior)

            # Px
            px_logit = self.decoder['px_nn'](z_prior_sample)
            xm = self.decoder['xm_layer'](px_logit)
            xv = torch.nn.functional.softplus(self.decoder['xv_layer'](px_logit))
            noise = torch.randn_like(torch.sqrt(xv))
            x_sample = xm + noise * xv

            xm_list.append(xm)
            xv_list.append(xv)
            x_list.append(x_sample)

        return (
            data, qy_logit, xm_list, xv_list,
            z_list, zm_list, zv_list,
            zm_prior_list, zv_prior_list
        )

    def evaluate_model(self, batch, batch_idx):
        """Evaluate the model on the data, computing average loss."""

        x = batch['data']

        if self.norm_in is not None:
            data = self.norm_in(x)

        qy_logit = self.encoder['qy_nn'](data)
        qy = torch.softmax(qy_logit, dim=1)

        y_ = torch.zeros([data.shape[0], self.k]).to(data.device)

        zm_list, zv_list, z_list = [], [], []
        xm_list, xv_list, x_list = [], [], []
        zm_prior_list, zv_prior_list = [], []

        for i in range(self.k):
            # One-hot y
            y = y_ + torch.eye(self.k).to(data.device)[i]

            # Qz
            h0 = self.encoder['y_transform'](y)
            xy = torch.cat([data, h0], dim=1)
            qz_logit = self.encoder['qz_nn'](xy)
            zm = self.encoder['zm_layer'](qz_logit)
            zv = torch.nn.functional.softplus(self.encoder['zv_layer'](qz_logit))
            noise = torch.randn_like(torch.sqrt(zv))
            z_sample = zm + noise * zv

            zm_list.append(zm)
            zv_list.append(zv)
            z_list.append(z_sample)

            # Pz (prior)
            pz_logit = self.decoder['pz_nn'](y)
            zm_prior = self.decoder['zm_prior_layer'](pz_logit)
            zv_prior = torch.nn.functional.softplus(self.decoder['zv_prior_layer'](pz_logit))
            noise = torch.randn_like(torch.sqrt(zv_prior))
            z_prior_sample = zm_prior + noise * zv_prior

            zm_prior_list.append(zm_prior)
            zv_prior_list.append(zv_prior)

            # Px
            px_logit = self.decoder['px_nn'](z_prior_sample)
            xm = self.decoder['xm_layer'](px_logit)
            xv = torch.nn.functional.softplus(self.decoder['xv_layer'](px_logit))
            noise = torch.randn_like(torch.sqrt(xv))
            x_sample = xm + noise * xv

            xm_list.append(xm)
            xv_list.append(xv)
            x_list.append(x_sample)

        xhat = torch.sum(qy.unsqueeze(-1) * torch.stack(x_list, dim=1), dim=1)

        if self.norm_in is not None:
            xhat = self.norm_in.inverse(xhat)

        z = torch.sum(qy.unsqueeze(-1) * torch.stack(z_list, dim=1), dim=1)

        return xhat, z, qy

    def decode(self, z):
        """
        Reconstruct x' from aggregated z
        """
        if z.dim() == 1:
            z = z.unsqueeze(0)

        px_logit = self.decoder['px_nn'](z)
        xm = self.decoder['xm_layer'](px_logit)
        xv = torch.nn.functional.softplus(self.decoder['xv_layer'](px_logit))
        noise = torch.randn_like(torch.sqrt(xv))
        x = xm + noise * xv

        if self.norm_in is not None:
            x = self.norm_in.inverse(x)

        return x

    def forward_cv(self, x):

        if self.norm_in is not None:
            x = self.norm_in(x)

        qy_logit = self.encoder['qy_nn'](x)
        qy = torch.softmax(qy_logit, dim=1)

        y_ = torch.zeros([x.shape[0], self.k]).to(x.device)

        zm_list, zv_list, z_list = [], [], []

        for i in range(self.k):
            # One-hot y
            y = y_ + torch.eye(self.k).to(x.device)[i]

            # Qz
            h0 = self.encoder['y_transform'](y)
            xy = torch.cat([x, h0], dim=1)
            qz_logit = self.encoder['qz_nn'](xy)
            zm = self.encoder['zm_layer'](qz_logit)
            zv = torch.nn.functional.softplus(self.encoder['zv_layer'](qz_logit))
            noise = torch.randn_like(torch.sqrt(zv))
            z_sample = zm + noise * zv

            zm_list.append(zm)
            zv_list.append(zv)
            z_list.append(z_sample)

        Z = torch.stack(z_list, dim=1)
        a = torch.sum(qy.unsqueeze(-1) * Z, dim=1)

        return a

    def training_step(self, train_batch, batch_idx):

        x = train_batch["data"]

        if "target" in train_batch:
            x_ref = train_batch["target"]
        else:
            x_ref = x

        data, qy_logit, xm_list, xv_list, z_list, zm_list, zv_list, zm_prior_list, zv_prior_list = self.encode_decode(x_ref)

        batch_loss, nent = self.loss_fn(data,
                                        qy_logit,
                                        xm_list, xv_list,
                                        z_list, zm_list, zv_list,
                                        zm_prior_list, zv_prior_list)

        loss = batch_loss.mean()
        ce_loss = nent.mean()

        name = "train" if self.training else "valid"
        self.log(f"{name}_loss", loss, on_epoch=True, on_step=True, prog_bar=True, logger=True)
        self.log(f"{name}_cross_entropy", ce_loss, on_step=True, on_epoch=True, prog_bar=True, logger=True)

        return loss

    def get_decoder(self, return_normalization=False):
        """Return a torch model with the decoder and optionally the normalization inverse"""
        if return_normalization:
            if self.norm_in is not None:
                inv_norm = Inverse(module=self.norm_in)
                decoder_model = torch.nn.Sequential(*[self.decoder, inv_norm])
            else:
                raise ValueError(
                    "return_normalization is set to True but self.norm_in is None"
                )
        else:
            decoder_model = self.decoder

        return decoder_model


# # Example of usage:

# # Define dimensions
# n_features = 1551   # Input dimension
# n_clusters = 5      # Output dimension for Qy
# n_cvs = 3           # Latent dimension (CVs)
# r_nent = 0.5        # Weight for the entropy regularization term.

# # Encoder sizes
# qy_dims = [32]
# qz_dims = [16, 16]

# # Decoder sizes
# pz_dims = [16, 16]
# px_dims = [128]

# options = {
#     "norm_in": {
#         "mode": "mean_std"
#     },
#     "optimizer": {
#         "lr": 1e-4
#     }
# }

# # Instantiate your GMVAECV
# model = GaussianMixtureVariationalAutoEncoder(k=n_clusters,
#                                                 n_cvs=n_cvs,
#                                                 n_features=n_features,
#                                                 r_nent=r_nent,
#                                                 qy_dims=qy_dims,
#                                                 qz_dims=qz_dims,
#                                                 pz_dims=pz_dims,
#                                                 px_dims=px_dims,
#                                                 options=options)
